"""
====================================================
Bootstrap and Closest Peak Direction Getters Example
====================================================

This example shows how choices in direction-getter impact fiber
tracking results by demonstrating the bootstrap direction getter (a type of
probabilistic tracking, as described in :footcite:p:`Berman2008` and the closest
peak direction getter (a type of deterministic tracking)
:footcite:p:`Amirbekian2016`.

This example is an extension of the
:ref:`sphx_glr_examples_built_quick_start_tracking_introduction_eudx.py`
example. Let's start by loading the necessary modules for executing this
tutorial.
"""

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, small_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import bootstrap_tracking, closestpeak_tracking
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)


seed_mask = labels == 2
white_matter = (labels == 1) | (labels == 2)
seeds = utils.seeds_from_mask(seed_mask, affine, density=1)

###############################################################################
# Next, we fit the CSD model.

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
csd_fit = csd_model.fit(data, mask=white_matter)

###############################################################################
# we use the CSA fit to calculate GFA, which will serve as our stopping
# criterion.

csa_model = CsaOdfModel(gtab, sh_order_max=6)
gfa = csa_model.fit(data, mask=white_matter).gfa
stopping_criterion = ThresholdStoppingCriterion(gfa, 0.25)

###############################################################################
# Next, we need to set up our two direction getters
#
#
# Example #1: Bootstrap direction getter with CSD Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
boot_streamline_generator = bootstrap_tracking(
    seeds,
    stopping_criterion,
    affine,
    step_size=0.5,
    data=data,
    model=csd_model,
    max_angle=30.0,
    sphere=small_sphere,
)
streamlines = Streamlines(boot_streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_bootstrap_dg.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(scene=scene, out_path="tractogram_bootstrap_dg.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum Bootstrap Probabilistic Direction Getter
#
#
# We have created a bootstrapped probabilistic set of streamlines. If you
# repeat the fiber tracking (keeping all inputs the same) you will NOT get
# exactly the same set of streamlines.
#
#
#
# Example #2: Closest peak direction getter with CSD Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pmf = csd_fit.odf(small_sphere).clip(min=0)
peak_streamline_generator = closestpeak_tracking(
    seeds,
    stopping_criterion,
    affine,
    step_size=0.5,
    sf=pmf,
    max_angle=30.0,
    sphere=small_sphere,
)
streamlines = Streamlines(peak_streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "closest_peak_dg_CSD.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_closest_peak_dg.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum Closest Peak Deterministic Direction Getter
#
#
# We have created a set of streamlines using the closest peak direction getter,
# which is a type of deterministic tracking. If you repeat the fiber tracking
# (keeping all inputs the same) you will get exactly the same set of
# streamlines.
#
#
#
# References
# ----------
#
# .. footbibliography::
#



from numpy.testing import assert_equal
import pytest

from dipy.align.streamwarp import (
    bundlewarp,
    bundlewarp_shape_analysis,
    bundlewarp_vector_filed,
)
from dipy.data import two_cingulum_bundles
from dipy.tracking.streamline import Streamlines, set_number_of_points
from dipy.utils.optpkg import optional_package

pd, have_pd, _ = optional_package("pandas")
needs_pandas = pytest.mark.skipif(not have_pd, reason="Requires pandas")


@needs_pandas
def test_bundlewarp():
    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, nb_points=20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, nb_points=20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    assert_equal(len(affine_bundle), len(cb2))
    assert_equal(len(deformed_bundle), len(cb2))
    assert_equal(len(deformed_bundle), len(affine_bundle))

    assert_equal(dists.shape, (len(cb2), len(cb1)))

    assert_equal(len(cb2), len(mp))

    assert_equal(len(cb2), len(warp))


@needs_pandas
def test_bundlewarp_vector_filed():
    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, nb_points=20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, nb_points=20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    offsets, directions, colors = bundlewarp_vector_filed(
        affine_bundle, deformed_bundle
    )

    assert_equal(len(offsets), len(cb2.get_data()))
    assert_equal(len(directions), len(cb2.get_data()))
    assert_equal(len(colors), len(cb2.get_data()))

    assert_equal(len(offsets), len(deformed_bundle.get_data()))
    assert_equal(len(directions), len(deformed_bundle.get_data()))
    assert_equal(len(colors), len(deformed_bundle.get_data()))


@needs_pandas
def test_bundle_shape_profile():
    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, nb_points=20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, nb_points=20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    n = 10
    shape_profile, stdv = bundlewarp_shape_analysis(
        affine_bundle, deformed_bundle, no_disks=n
    )

    assert_equal(len(shape_profile), n)
    assert_equal(len(stdv), n)

    n = 100
    shape_profile, stdv = bundlewarp_shape_analysis(
        affine_bundle, deformed_bundle, no_disks=n
    )

    assert_equal(len(shape_profile), n)
    assert_equal(len(stdv), n)


@needs_pandas
def test_transformation_dimensions():
    cingulum_bundles = two_cingulum_bundles()

    n = 20
    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, n)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, n)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    assert_equal(warp["gaussian_kernel"][0].shape, (n, n))
    assert_equal(warp["transforms"][0].shape, (n, 3))

    assert_equal(warp.columns.get_loc("gaussian_kernel"), 0)
    assert_equal(warp.columns.get_loc("transforms"), 1)



from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
from numpy.testing import assert_equal

from dipy.core.gradients import generate_bvecs, gradient_table
from dipy.io.image import load_nifti_data, save_nifti
from dipy.sims.voxel import multi_tensor
from dipy.workflows.reconst import ReconstIvimFlow


def test_reconst_ivim():
    with TemporaryDirectory() as out_dir:
        bvals = np.array(
            [
                0.0,
                10.0,
                20.0,
                30.0,
                40.0,
                60.0,
                80.0,
                100.0,
                120.0,
                140.0,
                160.0,
                180.0,
                200.0,
                300.0,
                400.0,
                500.0,
                600.0,
                700.0,
                800.0,
                900.0,
                1000.0,
            ]
        )
        N = len(bvals)
        bvecs = generate_bvecs(N)
        temp_bval_path = pjoin(out_dir, "temp.bval")
        np.savetxt(temp_bval_path, bvals)
        temp_bvec_path = pjoin(out_dir, "temp.bvec")
        np.savetxt(temp_bvec_path, bvecs)

        gtab = gradient_table(bvals, bvecs=bvecs)

        S0, f, D_star, D = 1000.0, 0.132, 0.00885, 0.000921

        mevals = np.array(([D_star, D_star, D_star], [D, D, D]))
        # This gives an isotropic signal.
        data = multi_tensor(
            gtab, mevals, snr=None, S0=S0, fractions=[f * 100, 100 * (1 - f)]
        )
        # Single voxel data
        data_single = data[0]
        temp_affine = np.eye(4)

        data_multi = np.zeros((2, 2, 1, len(gtab.bvals)), dtype=int)
        data_multi[0, 0, 0] = data_multi[0, 1, 0] = data_multi[1, 0, 0] = data_multi[
            1, 1, 0
        ] = data_single
        data_path = pjoin(out_dir, "tmp_data.nii.gz")
        save_nifti(data_path, data_multi, temp_affine, dtype=data_multi.dtype)

        mask = np.ones_like(data_multi[..., 0], dtype=np.uint8)
        mask_path = pjoin(out_dir, "tmp_mask.nii.gz")
        save_nifti(mask_path, mask, temp_affine)

        ivim_flow = ReconstIvimFlow()

        args = [data_path, temp_bval_path, temp_bvec_path, mask_path]

        msg = "Bounds for this fit have been set from experiments and * "
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=msg, category=UserWarning)
            ivim_flow.run(*args, out_dir=out_dir)

        S0_path = ivim_flow.last_generated_outputs["out_S0_predicted"]
        S0_data = load_nifti_data(S0_path)
        assert_equal(S0_data.shape, data_multi.shape[:-1])

        f_path = ivim_flow.last_generated_outputs["out_perfusion_fraction"]
        f_data = load_nifti_data(f_path)
        assert_equal(f_data.shape, data_multi.shape[:-1])

        D_star_path = ivim_flow.last_generated_outputs["out_D_star"]
        D_star_data = load_nifti_data(D_star_path)
        assert_equal(D_star_data.shape, data_multi.shape[:-1])

        D_path = ivim_flow.last_generated_outputs["out_D"]
        D_data = load_nifti_data(D_path)
        assert_equal(D_data.shape, data_multi.shape[:-1])



"""

Note
----

This file is copied (possibly with major modifications) from the
sources of the pycpd project - https://github.com/siavashk/pycpd.
It remains licensed as the rest of PyCPD (MIT license as of October 2010).

# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the PyCPD package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

import numbers
from warnings import warn

import numpy as np

from dipy.testing.decorators import warning_for_keywords


def gaussian_kernel(X, beta, Y=None):
    if Y is None:
        Y = X
    diff = X[:, None, :] - Y[None, :, :]
    diff = np.square(diff)
    diff = np.sum(diff, 2)
    return np.exp(-diff / (2 * beta**2))


def low_rank_eigen(G, num_eig):
    """Calculate num_eig eigenvectors and eigenvalues of gaussian matrix G.

    Enables lower dimensional solving.

    """
    S, Q = np.linalg.eigh(G)
    eig_indices = list(np.argsort(np.abs(S))[::-1][:num_eig])
    Q = Q[:, eig_indices]  # eigenvectors
    S = S[eig_indices]  # eigenvalues.
    return Q, S


def initialize_sigma2(X, Y):
    """Initialize the variance (sigma2).

    Parameters
    ----------
    X: numpy array
        NxD array of points for target.

    Y: numpy array
        MxD array of points for source.

    Returns
    -------
    sigma2: float
        Initial variance.
    """
    (N, D) = X.shape
    (M, _) = Y.shape
    diff = X[None, :, :] - Y[:, None, :]
    err = diff**2
    return np.sum(err) / (D * M * N)


@warning_for_keywords()
def lowrankQS(G, beta, num_eig, *, eig_fgt=False):
    """Calculate eigenvectors and eigenvalues of gaussian matrix G.

    !!!
    This function is a placeholder for implementing the fast
    gauss transform. It is not yet implemented.
    !!!

    Parameters
    ----------
    G: numpy array
        Gaussian kernel matrix.

    beta: float
        Width of the Gaussian kernel.

    num_eig: int
        Number of eigenvectors to use in lowrank calculation of G

    eig_fgt: bool
        If True, use fast gauss transform method to speed up.

    """
    # if we do not use FGT we construct affinity matrix G and find the
    # first eigenvectors/values directly

    if eig_fgt is False:
        S, Q = np.linalg.eigh(G)
        eig_indices = list(np.argsort(np.abs(S))[::-1][:num_eig])
        Q = Q[:, eig_indices]  # eigenvectors
        S = S[eig_indices]  # eigenvalues.

        return Q, S

    elif eig_fgt is True:
        raise Exception("Fast Gauss Transform Not Implemented!")


class DeformableRegistration:
    """
    Deformable point cloud registration.

    Attributes
    ----------
    X: numpy array
        NxD array of target points.

    Y: numpy array
        MxD array of source points.

    TY: numpy array
        MxD array of transformed source points.

    sigma2: float (positive)
        Initial variance of the Gaussian mixture model.

    N: int
        Number of target points.

    M: int
        Number of source points.

    D: int
        Dimensionality of source and target points

    iteration: int
        The current iteration throughout registration.

    max_iterations: int
        Registration will terminate once the algorithm has taken this
        many iterations.

    tolerance: float (positive)
        Registration will terminate once the difference between
        consecutive objective function values falls within this tolerance.

    w: float (between 0 and 1)
        Contribution of the uniform distribution to account for outliers.
        Valid values span 0 (inclusive) and 1 (exclusive).

    q: float
        The objective function value that represents the misalignment between
        source and target point clouds.

    diff: float (positive)
        The absolute difference between the current and previous objective
        function values.

    P: numpy array
        MxN array of probabilities.
        P[m, n] represents the probability that the m-th source point
        corresponds to the n-th target point.

    Pt1: numpy array
        Nx1 column array. Multiplication result between the transpose of P
        and a column vector of all 1s.

    P1: numpy array
        Mx1 column array.
        Multiplication result between P and a column vector of all 1s.

    Np: float (positive)
        The sum of all elements in P.

    alpha: float (positive)
        Represents the trade-off between the goodness of maximum likelihoo
        fit and regularization.

    beta: float(positive)
        Width of the Gaussian kernel.

    low_rank: bool
        Whether to use low rank approximation.

    num_eig: int
        Number of eigenvectors to use in lowrank calculation.
    """

    def __init__(
        self,
        X,
        Y,
        *args,
        sigma2=None,
        alpha=None,
        beta=None,
        low_rank=False,
        num_eig=100,
        max_iterations=None,
        tolerance=None,
        w=None,
        **kwargs,
    ):
        if not isinstance(X, np.ndarray) or X.ndim != 2:
            raise ValueError("The target point cloud (X) must be at a 2D numpy array.")

        if not isinstance(Y, np.ndarray) or Y.ndim != 2:
            raise ValueError("The source point cloud (Y) must be a 2D numpy array.")

        if X.shape[1] != Y.shape[1]:
            msg = "Both point clouds need to have the same number "
            msg += "of dimensions."
            raise ValueError(msg)

        if sigma2 is not None and (
            not isinstance(sigma2, numbers.Number) or sigma2 <= 0
        ):
            msg = f"Expected a positive value for sigma2 instead got: {sigma2}"
            raise ValueError(msg)

        if max_iterations is not None and (
            not isinstance(max_iterations, numbers.Number) or max_iterations < 0
        ):
            msg = "Expected a positive integer for max_iterations "
            msg += f"instead got: {max_iterations}"
            raise ValueError(msg)
        elif isinstance(max_iterations, numbers.Number) and not isinstance(
            max_iterations, int
        ):
            msg = "Received a non-integer value for max_iterations: "
            msg += f"{max_iterations}. Casting to integer."
            warn(msg, stacklevel=2)
            max_iterations = int(max_iterations)

        if tolerance is not None and (
            not isinstance(tolerance, numbers.Number) or tolerance < 0
        ):
            msg = "Expected a positive float for tolerance "
            msg += f"instead got: {tolerance}"
            raise ValueError(msg)

        if w is not None and (not isinstance(w, numbers.Number) or w < 0 or w >= 1):
            msg = "Expected a value between 0 (inclusive) and 1 (exclusive) "
            msg += f"for w instead got: {w}"
            raise ValueError(msg)

        self.X = X
        self.Y = Y
        self.TY = Y
        self.sigma2 = initialize_sigma2(X, Y) if sigma2 is None else sigma2
        (self.N, self.D) = self.X.shape
        (self.M, _) = self.Y.shape
        self.tolerance = 0.001 if tolerance is None else tolerance
        self.w = 0.0 if w is None else w
        self.max_iterations = 100 if max_iterations is None else max_iterations
        self.iteration = 0
        self.diff = np.inf
        self.q = np.inf
        self.P = np.zeros((self.M, self.N))
        self.Pt1 = np.zeros((self.N,))
        self.P1 = np.zeros((self.M,))
        self.PX = np.zeros((self.M, self.D))
        self.Np = 0

        if alpha is not None and (not isinstance(alpha, numbers.Number) or alpha <= 0):
            msg = "Expected a positive value for regularization parameter "
            msg += f"alpha. Instead got: {alpha}"
            raise ValueError(msg)

        if beta is not None and (not isinstance(beta, numbers.Number) or beta <= 0):
            msg = "Expected a positive value for the width of the coherent "
            msg += f"Gaussian kernel. Instead got: {beta}"

        self.alpha = 2 if alpha is None else alpha
        self.beta = 2 if beta is None else beta
        self.W = np.zeros((self.M, self.D))
        self.G = gaussian_kernel(self.Y, self.beta)
        self.low_rank = low_rank
        self.num_eig = num_eig
        if self.low_rank is True:
            self.Q, self.S = low_rank_eigen(self.G, self.num_eig)
            self.inv_S = np.diag(1.0 / self.S)
            self.S = np.diag(self.S)
            self.E = 0.0

    @warning_for_keywords()
    def register(self, *, callback=lambda **kwargs: None):
        """
        Perform the EM registration.

        Parameters
        ----------
        callback: function
            A function that will be called after each iteration.
            Can be used to visualize the registration process.

        Returns
        -------
        self.TY: numpy array
            MxD array of transformed source points.

        registration_parameters:
            Returned params dependent on registration method used.
        """
        self.transform_point_cloud()
        while self.iteration < self.max_iterations and self.diff > self.tolerance:
            self.iterate()
            if callable(callback):
                kwargs = {
                    "iteration": self.iteration,
                    "error": self.q,
                    "X": self.X,
                    "Y": self.TY,
                }
                callback(**kwargs)

        return self.TY, self.get_registration_parameters()

    def update_transform(self):
        """
        Calculate a new estimate of the deformable transformation.
        See Eq. 22 of https://arxiv.org/pdf/0905.2635.pdf.

        """
        if self.low_rank is False:
            A = np.dot(np.diag(self.P1), self.G) + self.alpha * self.sigma2 * np.eye(
                self.M
            )
            B = self.PX - np.dot(np.diag(self.P1), self.Y)
            self.W = np.linalg.solve(A, B)

        elif self.low_rank is True:
            # Matlab code equivalent can be found here:
            # https://github.com/markeroon/matlab-computer-vision-routines/tree/master/third_party/CoherentPointDrift
            dP = np.diag(self.P1)
            dPQ = np.matmul(dP, self.Q)
            F = self.PX - np.matmul(dP, self.Y)

            self.W = (
                1
                / (self.alpha * self.sigma2)
                * (
                    F
                    - np.matmul(
                        dPQ,
                        (
                            np.linalg.solve(
                                (
                                    self.alpha * self.sigma2 * self.inv_S
                                    + np.matmul(self.Q.T, dPQ)
                                ),
                                (np.matmul(self.Q.T, F)),
                            )
                        ),
                    )
                )
            )
            QtW = np.matmul(self.Q.T, self.W)
            self.E = self.E + self.alpha / 2 * np.trace(
                np.matmul(QtW.T, np.matmul(self.S, QtW))
            )

    @warning_for_keywords()
    def transform_point_cloud(self, *, Y=None):
        """Update a point cloud using the new estimate of the deformable
        transformation.

        Parameters
        ----------
        Y: numpy array, optional
            Array of points to transform - use to predict on new set of points.
            Best for predicting on new points not used to run initial
            registration. If None, self.Y used.

        Returns
        -------
        If Y is None, returns None.
        Otherwise, returns the transformed Y.


        """
        if Y is not None:
            G = gaussian_kernel(X=Y, beta=self.beta, Y=self.Y)
            return Y + np.dot(G, self.W)
        else:
            if self.low_rank is False:
                self.TY = self.Y + np.dot(self.G, self.W)

            elif self.low_rank is True:
                self.TY = self.Y + np.matmul(
                    self.Q, np.matmul(self.S, np.matmul(self.Q.T, self.W))
                )
                return

    def update_variance(self):
        """Update the variance of the mixture model.

        This is using the new estimate of the deformable transformation.
        See the update rule for sigma2 in
        Eq. 23 of of https://arxiv.org/pdf/0905.2635.pdf.

        """
        qprev = self.sigma2

        # The original CPD paper does not explicitly calculate the objective
        # functional. This functional will include terms from both the negative
        # log-likelihood and the Gaussian kernel used for regularization.
        self.q = np.inf

        xPx = np.dot(
            np.transpose(self.Pt1), np.sum(np.multiply(self.X, self.X), axis=1)
        )
        yPy = np.dot(
            np.transpose(self.P1), np.sum(np.multiply(self.TY, self.TY), axis=1)
        )
        trPXY = np.sum(np.multiply(self.TY, self.PX))

        self.sigma2 = (xPx - 2 * trPXY + yPy) / (self.Np * self.D)

        if self.sigma2 <= 0:
            self.sigma2 = self.tolerance / 10

        # Here we use the difference between the current and previous
        # estimate of the variance as a proxy to test for convergence.
        self.diff = np.abs(self.sigma2 - qprev)

    def get_registration_parameters(self):
        """Return the current estimate of the deformable transformation
        parameters.

        Returns
        -------
        self.G: numpy array
            Gaussian kernel matrix.

        self.W: numpy array
            Deformable transformation matrix.
        """
        return self.G, self.W

    def iterate(self):
        """Perform one iteration of the EM algorithm."""
        self.expectation()
        self.maximization()
        self.iteration += 1

    def expectation(self):
        """Compute the expectation step of the EM algorithm."""
        # (M, N)
        P = np.sum((self.X[None, :, :] - self.TY[:, None, :]) ** 2, axis=2)
        P = np.exp(-P / (2 * self.sigma2))
        c = (
            (2 * np.pi * self.sigma2) ** (self.D / 2)
            * self.w
            / (1.0 - self.w)
            * self.M
            / self.N
        )

        den = np.sum(P, axis=0, keepdims=True)  # (1, N)
        den = np.clip(den, np.finfo(self.X.dtype).eps, None) + c

        self.P = np.divide(P, den)
        self.Pt1 = np.sum(self.P, axis=0)
        self.P1 = np.sum(self.P, axis=1)
        self.Np = np.sum(self.P1)
        self.PX = np.matmul(self.P, self.X)

    def maximization(self):
        """Compute the maximization step of the EM algorithm."""
        self.update_transform()
        self.transform_point_cloud()
        self.update_variance()



cimport numpy as cnp

from dipy.direction.pmf cimport PmfGen
from dipy.tracking.tracker_parameters cimport TrackerParameters, TrackerStatus
from dipy.utils.fast_numpy cimport RNGState


cdef TrackerStatus deterministic_propagator(double* point,
                                            double* direction,
                                            TrackerParameters params,
                                            double* stream_data,
                                            PmfGen pmf_gen,
                                            RNGState* rng) noexcept nogil


cdef TrackerStatus probabilistic_propagator(double* point,
                                            double* direction,
                                            TrackerParameters params,
                                            double* stream_data,
                                            PmfGen pmf_gen,
                                            RNGState* rng) noexcept nogil

cdef TrackerStatus parallel_transport_propagator(double* point,
                                                 double* direction,
                                                 TrackerParameters params,
                                                 double* stream_data,
                                                 PmfGen pmf_gen,
                                                 RNGState* rng) noexcept nogil



cdef cnp.npy_intp _propagation_direction(double *point, double* prev, double* qa,
                                double *ind, double *odf_vertices,
                                double qa_thr, double ang_thr,
                                cnp.npy_intp *qa_shape,cnp.npy_intp* strides,
                                double *direction,double total_weight) noexcept nogil



# cython: boundscheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False


from dipy.align.fused_types cimport floating


cdef inline floating f_max(floating a, floating b) noexcept nogil:
    """Return the maximum of a and b."""
    return a if a >= b else b


cdef inline floating f_min(floating a, floating b) noexcept nogil:
    """Return the minimum of a and b."""
    return a if a <= b else b


cdef floating f_array_min(floating* arr, int n) noexcept nogil
cdef floating f_array_max(floating* arr, int n) noexcept nogil



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False

"""
Implementation of the Parallel Transport Tractography (PTT) algorithm by
:footcite:t:`Aydogan2021`. PTT Default parameter values are slightly different
than in Trekker to optimise performances. The rejection sampling
algorithm also uses fewer samples to estimate the maximum of the posterior, and
fewer tries to obtain a suitable propagation candidate. Moreover, the initial
tangent direction in this implementation is always obtained from the voxel-wise
peaks.

References
----------
.. footbibliography::
"""

cimport numpy as cnp
from libc.math cimport M_PI, pow, sin, cos, fabs
from libc.stdlib cimport malloc, free

from dipy.direction.probabilistic_direction_getter cimport \
        ProbabilisticDirectionGetter
from dipy.utils.fast_numpy cimport (copy_point, cross, normalize, random,
                                    random_perpendicular_vector,
                                    random_point_within_circle)
from dipy.tracking.stopping_criterion cimport (StreamlineStatus,
                                               StoppingCriterion,
                                               TRACKPOINT,
                                               ENDPOINT,
                                               OUTSIDEIMAGE,
                                               INVALIDPOINT)
from dipy.tracking.utils import min_radius_curvature_from_angle


cdef class PTTDirectionGetter(ProbabilisticDirectionGetter):
    """Parallel Transport Tractography (PTT) direction getter.
    """
    cdef double       angular_separation
    cdef double       data_support_exponent
    cdef double[3][3] frame
    cdef double       k1
    cdef double       k2
    cdef double       k_small
    cdef double       last_val
    cdef double       last_val_cand
    cdef double       max_angle
    cdef double       max_curvature
    cdef double[3]    position
    cdef int          probe_count
    cdef double       probe_length
    cdef double       probe_normalizer
    cdef int          probe_quality
    cdef double       probe_radius
    cdef double       probe_step_size
    cdef double[9]    propagator
    cdef double       step_size
    cdef int          rejection_sampling_max_try
    cdef int          rejection_sampling_nbr_sample
    cdef double[3]    voxel_size
    cdef double[3]    inv_voxel_size


    def __init__(self, pmf_gen, max_angle, sphere, pmf_threshold=None,
                 double probe_length=0.5, double probe_radius=0,
                 int probe_quality=3, int probe_count=1,
                 double data_support_exponent=1, **kwargs):
        """PTT used probe for estimating future propagation steps. A probe is a
        short, cylindrical model of the connecting segment.

        Parameters
        ----------
        pmf_gen : PmfGen
            Used to get probability mass function for selecting tracking
            directions.
        max_angle : float, [0, 90]
            Is used to set the upper limits for the k1 and k2 parameters
            of parallel transport frame (max_curvature).
        sphere : Sphere
            The set of directions to be used for tracking.
        pmf_threshold : float, [0., 1.]
            Used to remove direction from the probability mass function for
            selecting the tracking direction.
        probe_length : double
            The length of the probes. Shorter probe_length yields more
            dispersed fibers.
        probe_radius : double
            The radius of the probe. A large probe_radius helps mitigate noise
            in the pmf but it might make it harder to sample thin and intricate
            connections, also the boundary of fiber bundles might be eroded.
        probe_quality : integer,
            The quality of the probe. This parameter sets the number of
            segments to split the cylinder along the length of the
            probe (minimum=2).
        probe_count : integer
            The number of probes. This parameter sets the number of parallel
            lines used to model the cylinder (minimum=1).
        data_support_exponent : double
            Data support to the power dataSupportExponent is used for
            rejection sampling.

         """
        if not probe_length > 0:
            raise ValueError("probe_length must be greater than 0.")
        if not probe_radius >= 0:
            raise ValueError("probe_radius must be greater or equal to 0.")
        if not probe_quality >= 2:
            raise ValueError("probe_quality must be greater or equal than 2.")
        if not probe_count >= 1:
            raise ValueError("probe_count must be greater or equal than 1.")

        self.max_angle = max_angle
        self.probe_length = probe_length
        self.probe_radius = probe_radius
        self.probe_quality = probe_quality
        self.probe_count = probe_count
        self.probe_step_size = self.probe_length / (self.probe_quality - 1)
        self.probe_normalizer = 1.0 / float(self.probe_quality
                                            * self.probe_count)
        self.angular_separation = 2.0 * M_PI / float(self.probe_count)
        self.data_support_exponent = data_support_exponent

        self.k_small = 0.0001
        self.rejection_sampling_max_try = 100
        self.rejection_sampling_nbr_sample = 10 # Adaptively set in Trekker.

        ProbabilisticDirectionGetter.__init__(self, pmf_gen, max_angle, sphere,
                                       pmf_threshold=pmf_threshold, **kwargs)


    cdef void initialize_candidate(self, double[:] init_dir):
        """"Initialize the parallel transport frame.

        After initial position is set, a parallel transport frame is set using
        the initial direction (a walking frame, i.e., 3 orthonormal vectors,
        plus 2 scalars, i.e. k1 and k2).

        A point and parallel transport frame parametrizes a curve that is named
        the "probe". Using probe parameters (probe_length, probe_radius,
        probe_quality, probe_count), a short fiber bundle segment is modelled.

        Parameters
        ----------
        init_dir : np.array
            Initial tracking direction (tangent)

        """
        cdef double[3] position
        cdef int count
        cdef cnp.npy_intp i

        # Initialize Frame
        self.frame[0][0] = init_dir[0]
        self.frame[0][1] = init_dir[1]
        self.frame[0][2] = init_dir[2]
        random_perpendicular_vector(&self.frame[2][0], &self.frame[0][0])
        cross(&self.frame[1][0], &self.frame[2][0], &self.frame[0][0])
        self.k1, self.k2 = random_point_within_circle(self.max_curvature)

        self.last_val = 0

        if self.probe_count == 1:
            self.last_val = self.pmf_gen.get_pmf_value_c(self.position,
                                                         self.frame[0])
        else:
            for count in range(self.probe_count):
                for i in range(3):
                    position[i] = (self.position[i]
                                   + self.frame[1][i] * self.probe_radius
                                   * cos(count * self.angular_separation)
                                   * self.inv_voxel_size[i]
                                   +
                                   self.frame[2][i] * self.probe_radius
                                   * sin(count * self.angular_separation)
                                   * self.inv_voxel_size[i])

                self.last_val += self.pmf_gen.get_pmf_value_c(position,
                                                              self.frame[0])


    cdef void prepare_propagator(self, double arclength) nogil:
        """Prepare the propagator.

        The propagator used for transporting the moving frame forward.

        Parameters
        ----------
        arclength : double
            Arclenth, which is equivalent to step size along the arc

        """
        cdef double tmp_arclength

        if (fabs(self.k1) < self.k_small
            and fabs(self.k2) < self.k_small):
            self.propagator[0] = arclength
            self.propagator[1] = 0
            self.propagator[2] = 0
            self.propagator[3] = 1
            self.propagator[4] = 0
            self.propagator[5] = 0
            self.propagator[6] = 0
            self.propagator[7] = 0
            self.propagator[8] = 1
        else:
            if fabs(self.k1) < self.k_small:
                self.k1 = self.k_small
            if fabs(self.k2) < self.k_small:
                self.k2 = self.k_small

            tmp_arclength  = arclength * arclength / 2.0

            self.propagator[0] = arclength
            self.propagator[1] = self.k1 * tmp_arclength
            self.propagator[2] = self.k2 * tmp_arclength
            self.propagator[3] = (1
                                  - self.k2 * self.k2 * tmp_arclength
                                  - self.k1 * self.k1 * tmp_arclength)
            self.propagator[4] = self.k1 * arclength
            self.propagator[5] = self.k2 * arclength
            self.propagator[6] = -self.k2 * arclength
            self.propagator[7] = -self.k1 * self.k2 * tmp_arclength
            self.propagator[8] = (1 - self.k2 * self.k2 * tmp_arclength)


    cdef double calculate_data_support(self):
        """Calculates data support for the candidate probe."""

        cdef double fod_amp
        cdef double[3] position
        cdef double[3][3] frame
        cdef double[3] tangent
        cdef double[3] normal
        cdef double[3] binormal
        cdef double[3] new_position
        cdef double likelihood
        cdef int c, i, j, q

        self.prepare_propagator(self.probe_step_size)

        for i in range(3):
            position[i] = self.position[i]
            for j in range(3):
                frame[i][j] = self.frame[i][j]

        likelihood = self.last_val

        for q in range(1, self.probe_quality):
            for i in range(3):
                position[i] = \
                    (self.propagator[0] * frame[0][i] * self.voxel_size[i]
                     + self.propagator[1] * frame[1][i] * self.voxel_size[i]
                     + self.propagator[2] * frame[2][i] * self.voxel_size[i]
                     + position[i])
                tangent[i] = (self.propagator[3] * frame[0][i]
                              + self.propagator[4] * frame[1][i]
                              + self.propagator[5] * frame[2][i])
            normalize(&tangent[0])

            if q < (self.probe_quality - 1):
                for i in range(3):
                    binormal[i] = (self.propagator[6] * frame[0][i]
                                   + self.propagator[7] * frame[1][i]
                                   + self.propagator[8] * frame[2][i])
                cross(&normal[0], &binormal[0], &tangent[0])

                copy_point(&tangent[0], &frame[0][0])
                copy_point(&normal[0], &frame[1][0])
                copy_point(&binormal[0], &frame[2][0])

            if self.probe_count == 1:
                fod_amp = self.pmf_gen.get_pmf_value_c(position, tangent)
                fod_amp = fod_amp if fod_amp > self.pmf_threshold else 0
                self.last_val_cand = fod_amp
                likelihood += self.last_val_cand
            else:
                self.last_val_cand = 0
                if q == self.probe_quality-1:
                    for i in range(3):
                        binormal[i] = (self.propagator[6] * frame[0][i]
                                      + self.propagator[7] * frame[1][i]
                                      + self.propagator[8] * frame[2][i])
                    cross(&normal[0], &binormal[0], &tangent[0])

                for c in range(self.probe_count):
                    for i in range(3):
                        new_position[i] = (position[i]
                                           + normal[i] * self.probe_radius
                                           * cos(c * self.angular_separation)
                                           * self.inv_voxel_size[i]
                                           + binormal[i] * self.probe_radius
                                           * sin(c * self.angular_separation)
                                           * self.inv_voxel_size[i])
                    fod_amp = self.pmf_gen.get_pmf_value_c(new_position, tangent)
                    fod_amp = fod_amp if fod_amp > self.pmf_threshold else 0
                    self.last_val_cand += fod_amp

                likelihood += self.last_val_cand

        likelihood *= self.probe_normalizer
        if self.data_support_exponent != 1:
            likelihood = pow(likelihood, self.data_support_exponent)

        return likelihood


    cdef int initialize(self, double[:] seed_point, double[:] seed_direction):
        """Sample an initial curve by rejection sampling.

        Parameters
        ----------
        seed_point : double[3]
            Initial point
        seed_direction : double[3]
            Initial direction

        Returns
        -------
        status : int
            Returns 0 if the initialization was successful, or
            1 otherwise.
        """
        cdef double data_support = 0
        cdef double max_posterior = 0
        cdef int tries

        self.position[0] = seed_point[0]
        self.position[1] = seed_point[1]
        self.position[2] = seed_point[2]

        for tries in range(self.rejection_sampling_nbr_sample):
            self.initialize_candidate(seed_direction)
            data_support = self.calculate_data_support()
            if data_support > max_posterior:
                max_posterior = data_support

        # Compensation for underestimation of max posterior estimate
        max_posterior = pow(2.0 * max_posterior, self.data_support_exponent)

        # Initialization is successful if a suitable candidate can be sampled
        # within the trial limit
        for tries in range(self.rejection_sampling_max_try):
            self.initialize_candidate(seed_direction)
            if (random() * max_posterior <= self.calculate_data_support()):
                self.last_val = self.last_val_cand
                return 0

        return 1

    cdef int propagate(self):
        """Propagates the position by step_size amount. The propagation is using
        the parameters of the last candidate curve. Then, randomly generate
        curve parametrization from the current position. The walking frame
        is the same, only the k1 and k2 parameters are randomly picked.
        Rejection sampling is used to pick the next curve using the data
        support (likelihood).

        Returns
        -------
        status : int
            Returns 0 if the propagation was successful, or
            1 otherwise.
        """
        cdef double max_posterior = 0
        cdef double data_support = 0
        cdef double[3] tangent
        cdef int tries
        cdef cnp.npy_intp i

        self.prepare_propagator(self.step_size)

        for i in range(3):
            self.position[i] = \
                (self.propagator[0] * self.frame[0][i] * self.inv_voxel_size[i]
                + self.propagator[1] * self.frame[1][i] * self.inv_voxel_size[i]
                + self.propagator[2] * self.frame[2][i] * self.inv_voxel_size[i]
                + self.position[i])
            tangent[i] = (self.propagator[3] * self.frame[0][i]
                          + self.propagator[4] * self.frame[1][i]
                          + self.propagator[5] * self.frame[2][i])
            self.frame[2][i] = (self.propagator[6] * self.frame[0][i]
                                + self.propagator[7] * self.frame[1][i]
                                + self.propagator[8] * self.frame[2][i])
        normalize(&tangent[0])
        cross(&self.frame[1][0], &self.frame[2][0], &tangent[0])
        normalize(&self.frame[1][0])
        cross(&self.frame[2][0], &tangent[0], &self.frame[1][0])
        self.frame[0][0] = tangent[0]
        self.frame[0][1] = tangent[1]
        self.frame[0][2] = tangent[2]

        for tries in range(self.rejection_sampling_nbr_sample):
            self.k1, self.k2 = random_point_within_circle(self.max_curvature)
            data_support = self.calculate_data_support()
            if data_support > max_posterior:
                max_posterior = data_support

        # Compensation for underestimation of max posterior estimate
        max_posterior = pow(2.0 * max_posterior, self.data_support_exponent)

        # Propagation is successful if a suitable candidate can be sampled
        # within the trial limit
        for tries in range(self.rejection_sampling_max_try):
            self.k1, self.k2 = random_point_within_circle(self.max_curvature)
            if random() * max_posterior <= self.calculate_data_support():
                self.last_val = self.last_val_cand
                return 0

        return 1


    # @cython.boundscheck(False)
    # @cython.wraparound(False)
    # @cython.cdivision(True)
    cpdef tuple generate_streamline(self,
                                    double[::1] seed,
                                    double[::1] dir,
                                    double[::1] voxel_size,
                                    double step_size,
                                    StoppingCriterion stopping_criterion,
                                    cnp.float_t[:, :] streamline,
                                    StreamlineStatus stream_status,
                                    int fixedstep):
        cdef:
            cnp.npy_intp i
            cnp.npy_intp len_streamlines = streamline.shape[0]
            double average_voxel_size = 0

        if not fixedstep > 0:
           raise ValueError("PTT only supports fixed step size.")

        self.step_size = step_size
        for i in range(3):
            self.voxel_size[i] = voxel_size[i]
            self.inv_voxel_size[i] = 1 / voxel_size[i]
            average_voxel_size += voxel_size[i] / 3

        # convert max_angle from degrees to radians
        self.max_curvature = 1 / min_radius_curvature_from_angle(
            self.max_angle * M_PI / 180.0, self.step_size / average_voxel_size)

        copy_point(&seed[0], &streamline[0,0])
        i = 0
        stream_status = TRACKPOINT

        if not self.initialize(seed, dir):
            # the initialization was successful
            for i in range(1, len_streamlines):
                if self.propagate():
                    # the propagation failed
                    break
                copy_point(<double *>&self.position[0], &streamline[i, 0])
                stream_status = stopping_criterion\
                    .check_point_c(<double * > &self.position[0])
                if stream_status == TRACKPOINT:
                    continue
                elif (stream_status == ENDPOINT or
                    stream_status == INVALIDPOINT or
                    stream_status == OUTSIDEIMAGE):
                    break
            else:
                # maximum length has been reached, return everything
                i = streamline.shape[0]
        return i, stream_status




from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")


class SingleLayerPerceptron:
    @warning_for_keywords()
    def __init__(
        self,
        *,
        input_shape=(28, 28),
        num_hidden=128,
        act_hidden="relu",
        dropout=0.2,
        num_out=10,
        act_out="softmax",
        optimizer="adam",
        loss="sparse_categorical_crossentropy",
    ):
        """Single Layer Perceptron with Dropout.

        Parameters
        ----------
        input_shape : tuple
            Shape of data to be trained
        num_hidden : int
            Number of nodes in hidden layer
        act_hidden : string
            Activation function used in hidden layer
        dropout : float
            Dropout ratio
        num_out : 10
            Number of nodes in output layer
        act_out : string
            Activation function used in output layer
        optimizer :  string
            Select optimizer. Default adam.
        loss : string
            Select loss function for measuring accuracy.
            Default sparse_categorical_crossentropy.

        """
        self.accuracy = None
        self.loss = None

        model = tf.keras.models.Sequential(
            [
                tf.keras.layers.Input(shape=input_shape),
                tf.keras.layers.Flatten(),
                tf.keras.layers.Dense(num_hidden, activation=act_hidden),
                tf.keras.layers.Dropout(dropout),
                tf.keras.layers.Dense(num_out, activation=act_out),
            ]
        )

        model.compile(optimizer=optimizer, loss=loss, metrics=["accuracy"])

        self.model = model

    def summary(self):
        """Get the summary of the model.

        The summary is textual and includes information about:
        The layers and their order in the model.
        The output shape of each layer.

        Returns
        -------
        summary : NoneType
            the summary of the model

        """
        return self.model.summary()

    @warning_for_keywords()
    def fit(self, x_train, y_train, *, epochs=5):
        """Train the model on train dataset.

        The fit method will train the model for a fixed
        number of epochs (iterations) on a dataset.

        Parameters
        ----------
        x_train : ndarray
            the x_train is the train dataset
        y_train : ndarray shape=(BatchSize,)
            the y_train is the labels of the train dataset
        epochs : int (Default = 5)
            the number of epochs

        Returns
        -------
        hist : object
            A History object. Its History.history attribute is a record of
            training loss values and metrics values at successive epochs

        """
        hist = self.model.fit(x_train, y_train, epochs=epochs)
        self.accuracy = hist.history["accuracy"][0]
        self.loss = hist.history["loss"][0]
        return hist

    @warning_for_keywords()
    def evaluate(self, x_test, y_test, *, verbose=2):
        """Evaluate the model on test dataset.

        The evaluate method will evaluate the model on a test
        dataset.

        Parameters
        ----------
        x_test : ndarray
            the x_test is the test dataset
        y_test : ndarray shape=(BatchSize,)
            the y_test is the labels of the test dataset
        verbose : int (Default = 2)
            By setting verbose 0, 1 or 2 you just say how do you want to
            'see' the training progress for each epoch.

        Returns
        -------
        evaluate : List
            return list of loss value and accuracy value on test dataset

        """
        return self.model.evaluate(x_test, y_test, verbose=verbose)

    def predict(self, x_test):
        """Predict the output from input samples.

        The predict method will generates output predictions
        for the input samples.

        Parameters
        ----------
        x_train : ndarray
            the x_test is the test dataset or input samples

        Returns
        -------
        predict : ndarray shape(TestSize,OutputSize)
            Numpy array(s) of predictions.

        """
        return self.model.predict(x_test)


class MultipleLayerPercepton:
    @warning_for_keywords()
    def __init__(
        self,
        *,
        input_shape=(28, 28),
        num_hidden=(128,),
        act_hidden="relu",
        dropout=0.2,
        num_out=10,
        act_out="softmax",
        loss="sparse_categorical_crossentropy",
        optimizer="adam",
    ):
        """Multiple Layer Perceptron with Dropout.

        Parameters
        ----------
        input_shape : tuple
            Shape of data to be trained
        num_hidden : array-like
            List of number of nodes in hidden layers
        act_hidden : string
            Activation function used in hidden layer
        dropout : float
            Dropout ratio
        num_out : 10
            Number of nodes in output layer
        act_out : string
            Activation function used in output layer
        optimizer :  string
            Select optimizer. Default adam.
        loss : string
            Select loss function for measuring accuracy.
            Default sparse_categorical_crossentropy.

        """
        self.input_shape = input_shape
        self.num_hidden = num_hidden
        self.act_hidden = act_hidden
        self.dropout = dropout
        self.num_out = num_out
        self.act_out = act_out
        self.loss = loss
        self.optimizer = optimizer
        self.accuracy = None

        # model building

        inp = tf.keras.layers.Input(shape=self.input_shape)
        x = tf.keras.layers.Flatten()(inp)

        for i in range(len(self.num_hidden)):
            x = tf.keras.layers.Dense(self.num_hidden[i])(x)

        x = tf.keras.layers.Dropout(self.dropout)(x)
        out = tf.keras.layers.Dense(self.num_out, activation=self.act_out)(x)

        self.model = tf.keras.models.Model(inputs=inp, outputs=out)

        # compiling the model
        self.model.compile(
            optimizer=self.optimizer, loss=self.loss, metrics=["accuracy"]
        )

    def summary(self):
        """Get the summary of the model.

        The summary is textual and includes information about:
        The layers and their order in the model.
        The output shape of each layer.

        Returns
        -------
        summary :  NoneType
            the summary of the model

        """
        return self.model.summary()

    @warning_for_keywords()
    def fit(self, x_train, y_train, *, epochs=5):
        """Train the model on train dataset.

        The fit method will train the model for a fixed
        number of epochs (iterations) on a dataset.

        Parameters
        ----------
        x_train : ndarray
            the x_train is the train dataset
        y_train : ndarray shape=(BatchSize,)
            the y_train is the labels of the train dataset
        epochs : int (Default = 5)
            the number of epochs

        Returns
        -------
        hist : object
            A History object. Its History.history attribute is a record of
            training loss values and metrics values at successive epochs

        """
        hist = self.model.fit(x_train, y_train, epochs=epochs)
        self.accuracy = hist.history["accuracy"][0]
        self.loss = hist.history["loss"][0]
        return hist

    @warning_for_keywords()
    def evaluate(self, x_test, y_test, *, verbose=2):
        """Evaluate the model on test dataset.

        The evaluate method will evaluate the model on a test
        dataset.

        Parameters
        ----------
        x_test : ndarray
            the x_test is the test dataset
        y_test : ndarray shape=(BatchSize,)
            the y_test is the labels of the test dataset
        verbose : int (Default = 2)
            By setting verbose 0, 1 or 2 you just say how do you want to
            'see' the training progress for each epoch.

        Returns
        -------
        evaluate : List
            return list of loss value and accuracy value on test dataset

        """
        return self.model.evaluate(x_test, y_test, verbose=verbose)

    def predict(self, x_test):
        """Predict the output from input samples.

        The predict method will generates output predictions
        for the input samples.

        Parameters
        ----------
        x_train : ndarray
            the x_test is the test dataset or input samples

        Returns
        -------
        predict : ndarray shape(TestSize,OutputSize)
            Numpy array(s) of predictions.

        """
        return self.model.predict(x_test)



"""
Diffusion Imaging in Python
============================

For more information, please visit https://dipy.org

Subpackages
-----------
::

 align         -- Registration, streamline alignment, volume resampling
 core          -- Spheres, gradient tables
 core.geometry -- Spherical geometry, coordinate and vector manipulation
 core.meshes   -- Point distributions on the sphere
 data          -- Small testing datasets
 denoise       -- Denoising algorithms
 direction     -- Manage peaks and tracking
 io            -- Loading/saving of dpy datasets
 nn            -- Neural networks algorithms
 reconst       -- Signal reconstruction modules (tensor, spherical harmonics,
                  diffusion spectrum, etc.)
 segment       -- Tractography segmentation
 sims          -- MRI phantom signal simulation
 stats         -- Tractometry
 tracking      -- Tractography, metrics for streamlines
 viz           -- Visualization and GUIs
 workflows      -- Predefined Command line for common tasks

Utilities
---------
::

 test          -- Run unittests
 __version__   -- Dipy version

"""

from dipy.version import version as __version__


def get_info():
    from os.path import dirname
    import sys

    import numpy

    import dipy

    return {
        "pkg_path": dirname(__file__),
        "commit_hash": dipy.version.git_revision,
        "sys_version": sys.version,
        "sys_executable": sys.executable,
        "sys_platform": sys.platform,
        "np_version": numpy.__version__,
        "dipy_version": dipy.__version__,
    }


submodules = [
    "align",
    "core",
    "data",
    "denoise",
    "direction",
    "io",
    "nn",
    "reconst",
    "segment",
    "sims",
    "stats",
    "tracking",
    "utils",
    "viz",
    "workflows",
    "tests",
    "testing",
]

__all__ = submodules + ["__version__", "setup_test", "get_info"]



"""
=================================================================
Estimating diffusion time dependent q-space indices using qt-dMRI
=================================================================
Effective representation of the four-dimensional diffusion MRI signal --
varying over three-dimensional q-space and diffusion time -- is a sought-after
and still unsolved challenge in diffusion MRI (dMRI). We propose a functional
basis approach that is specifically designed to represent the dMRI signal in
this qtau-space :footcite:p:`Fick2018`. Following recent terminology, we refer
to our qtau-functional basis as :math:`q\tau`-dMRI. We use GraphNet
regularization --imposing both signal smoothness and sparsity -- to drastically
reduce the number of diffusion-weighted images (DWIs) that is needed to
represent the dMRI signal in the qtau-space. As the main contribution,
:math:`q\tau`-dMRI provides the framework to -- without making biophysical
assumptions -- represent the :math:`q\tau`-space signal and estimate
time-dependent q-space indices (:math:`q\tau`-indices), providing a new means
for studying diffusion in nervous tissue. :math:`q\tau`-dMRI is the first of its
kind in being specifically designed to provide open interpretation of the
:math:`q\tau`-diffusion signal.

:math:`q\tau`-dMRI can be seen as a time-dependent extension of the MAP-MRI
functional basis :footcite:p:`Ozarslan2013`, and all the previously proposed
q-space can be estimated for any diffusion time. These include rotationally
invariant quantities such as the Mean Squared Displacement (MSD), Q-space
Inverse Variance (QIV) and Return-To-Origin Probability (RTOP). Also
directional indices such as the Return To the Axis Probability (RTAP) and
Return To the Plane Probability (RTPP) are available, as well as the
Orientation Distribution Function (ODF).

In this example we illustrate how to use the :math:`q\tau`-dMRI to estimate
time-dependent q-space indices from a :math:`q\tau`-acquisition of a mouse.

First import the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.data.fetcher import (
    fetch_qtdMRI_test_retest_2subjects,
    read_qtdMRI_test_retest_2subjects,
)
from dipy.reconst import dti, qtdmri

###############################################################################
# Download and read the data for this tutorial.
#
# :math:`q\tau`-dMRI requires data with multiple gradient directions, gradient
# strength and diffusion times. We will use the test-retest acquisitions of two
# mice that were used in the test-retest study by :footcite:p:`Fick2018`. The
# data itself is freely available and citeable at :footcite:p:`Wassermann2017`.

fetch_qtdMRI_test_retest_2subjects()
data, cc_masks, gtabs = read_qtdMRI_test_retest_2subjects()

###############################################################################
# data contains 4 qt-dMRI datasets of size [80, 160, 5, 515]. The first two are
# the test-retest datasets of the first mouse and the second two are those of
# the second mouse. cc_masks contains 4 corresponding binary masks for the
# corpus callosum voxels in the middle slice that were used in the test-retest
# study. Finally, gtab contains the qt-dMRI gradient tables for the DWIs in the
# dataset.
#
# The data consists of 515 DWIs, divided over 35 shells, with 7 "gradient
# strength shells" up to 491 mT/m, 5 equally spaced "pulse separation shells"
# (big_delta) between [10.8-20] ms and a pulse duration (small_delta) of 5ms.
#
# To visualize qt-dMRI acquisition schemes in an intuitive way, the qtdmri
# module provides a visualization function to illustrate the relationship
# between gradient strength (G), pulse separation (big_delta) and b-value:

plt.figure()
qtdmri.visualise_gradient_table_G_Delta_rainbow(gtabs[0])
plt.savefig("qt-dMRI_acquisition_scheme.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# In the figure the dots represent measured DWIs in any direction, for a given
# gradient strength and pulse separation. The background isolines represent the
# corresponding b-values for different combinations of G and big_delta.
#
#
# Next, we visualize the middle slices of the test-retest data sets with their
# corresponding masks. To better illustrate the white matter architecture in
# the data, we calculate DTI's fractional anisotropy (FA) over the whole slice
# and project the corpus callosum mask on the FA image.:

subplot_titles = [
    "Subject1 Test",
    "Subject1 Retest",
    "Subject2 Test",
    "Subject2 Retest",
]
fig = plt.figure()
plt.subplots(nrows=2, ncols=2)
for i, (data_, mask_, gtab_) in enumerate(zip(data, cc_masks, gtabs)):
    # take the middle slice
    data_middle_slice = data_[:, :, 2]
    mask_middle_slice = mask_[:, :, 2]

    # estimate fractional anisotropy (FA) for this slice
    tenmod = dti.TensorModel(gtab_)
    tenfit = tenmod.fit(data_middle_slice, mask=data_middle_slice[..., 0] > 0)
    fa = tenfit.fa

    # set mask color to green with 0.5 opacity as overlay
    mask_template = np.zeros(np.r_[mask_middle_slice.shape, 4])
    mask_template[mask_middle_slice == 1] = np.r_[0.0, 1.0, 0.0, 0.5]

    # produce the FA images with corpus callosum masks.
    plt.subplot(2, 2, 1 + i)
    plt.title(subplot_titles[i], fontsize=15)
    plt.imshow(fa, cmap="Greys_r", origin="lower", interpolation="nearest")
    plt.imshow(mask_template, origin="lower", interpolation="nearest")
    plt.axis("off")
plt.tight_layout()
plt.savefig("qt-dMRI_datasets_fa_with_ccmasks.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Next, we use qt-dMRI to estimate of time-dependent q-space indices
# (q$\tau$-indices) for the masked voxels in the corpus callosum of each
# dataset. In particular, we estimate the Return-to-Original, Return-to-Axis
# and Return-to-Plane Probability (RTOP, RTAP and RTPP), as well as the Mean
# Squared Displacement (MSD).
#
#
#
# In this example we don't extrapolate the data beyond the maximum diffusion
# time, so we estimate :math:`q\tau` indices between the minimum and maximum
# diffusion times of the data at 5 equally spaced points. However, it should
# the noted that qt-dMRI's combined smoothness and sparsity regularization
# allows for smooth interpolation at any :math:`q\tau` position. In other
# words, once the basis is fitted to the data, its coefficients describe the
# entire :math:`q\tau`-space, and any :math:`q\tau`-position can be freely
# recovered. This including points beyond the dataset's maximum
# :math:`q\tau` value (although this should be done with caution).

tau_min = gtabs[0].tau.min()
tau_max = gtabs[0].tau.max()
taus = np.linspace(tau_min, tau_max, 5)

qtdmri_fits = []
msds = []
rtops = []
rtaps = []
rtpps = []
for data_, mask_, gtab_ in zip(data, cc_masks, gtabs):
    # select the corpus callosum voxel for every dataset
    cc_voxels = data_[mask_ == 1]
    # initialize the qt-dMRI model.
    # recommended basis orders are radial_order=6 and time_order=2.
    # The combined Laplacian and l1-regularization using Generalized
    # Cross-Validation (GCV) and Cross-Validation (CV) settings is most robust,
    # but can be used separately and with weightings preset to any positive
    # value to optimize for speed.
    qtdmri_mod = qtdmri.QtdmriModel(
        gtab_,
        radial_order=6,
        time_order=2,
        laplacian_regularization=True,
        laplacian_weighting="GCV",
        l1_regularization=True,
        l1_weighting="CV",
    )
    # fit the model.
    # Here we take every 5th voxel for speed, but of course all voxels can be
    # fit for a more robust result later on.
    qtdmri_fit = qtdmri_mod.fit(cc_voxels[::5])
    qtdmri_fits.append(qtdmri_fit)
    # We estimate MSD, RTOP, RTAP and RTPP for the chosen diffusion times.
    msds.append(np.array(list(map(qtdmri_fit.msd, taus))))
    rtops.append(np.array(list(map(qtdmri_fit.rtop, taus))))
    rtaps.append(np.array(list(map(qtdmri_fit.rtap, taus))))
    rtpps.append(np.array(list(map(qtdmri_fit.rtpp, taus))))

###############################################################################
# The estimated :math:`q\tau`-indices, for the chosen diffusion times, are now
# stored in msds, rtops, rtaps and rtpps. The trends of these
# :math:`q\tau`-indices over time say something about the restriction of
# diffusing particles over time, which is currently a hot topic in the dMRI
# community. We evaluate the test-retest reproducibility for the two subjects
# by plotting the :math:`q\tau`-indices for each subject together. This
# example will produce similar results as Fig. 10 in :footcite:p:`Fick2018`.
#
# We first define a small function to plot the mean and standard deviation of
# the :math:`q\tau`-index trends in a subject.


def plot_mean_with_std(ax, time, ind1, plotcolor, ls="-", std_mult=1, label=""):
    means = np.mean(ind1, axis=1)
    stds = np.std(ind1, axis=1)
    ax.plot(time, means, c=plotcolor, lw=3, label=label, ls=ls)
    ax.fill_between(
        time,
        means + std_mult * stds,
        means - std_mult * stds,
        alpha=0.15,
        color=plotcolor,
    )
    ax.plot(time, means + std_mult * stds, alpha=0.25, color=plotcolor)
    ax.plot(time, means - std_mult * stds, alpha=0.25, color=plotcolor)


###############################################################################
# We start by showing the test-retest MSD of both subjects over time. We plot
# the :math:`q\tau`-indices together with :math:`q\tau`-index trends of free
# diffusion with different diffusivities as background.

# we first generate the data to produce the background index isolines.
Delta_ = np.linspace(0.005, 0.02, 100)
MSD_ = np.linspace(4e-5, 10e-5, 100)
Delta_grid, MSD_grid = np.meshgrid(Delta_, MSD_)
D_grid = MSD_grid / (6 * Delta_grid)
D_levels = np.r_[1, 5, 7, 10, 14, 23, 30] * 1e-4

fig = plt.figure(figsize=(10, 3))
# start with the plot of subject 1.
ax = plt.subplot(1, 2, 1)
# first plot the background
plt.contourf(Delta_ * 1e3, 1e5 * MSD_, D_grid, levels=D_levels, cmap="Greys", alpha=0.5)

# plot the test-retest mean MSD and standard deviation of subject 1.
plot_mean_with_std(ax, taus * 1e3, 1e5 * msds[0], "r", "dashdot", label="MSD Test")
plot_mean_with_std(ax, taus * 1e3, 1e5 * msds[1], "g", "dashdot", label="MSD Retest")
ax.legend(fontsize=13)
# plot some text markers to clarify the background diffusivity lines.
ax.text(0.0091 * 1e3, 6.33, "D=14e-4", fontsize=12, rotation=35)
ax.text(0.0091 * 1e3, 4.55, "D=10e-4", fontsize=12, rotation=25)
ax.set_ylim(4, 9.5)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_title(r"Test-Retest MSD($\tau$) Subject 1", fontsize=15)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_ylabel("MSD ($10^{-5}mm^2$)", fontsize=17)

# then do the same thing for subject 2.
ax = plt.subplot(1, 2, 2)
plt.contourf(Delta_ * 1e3, 1e5 * MSD_, D_grid, levels=D_levels, cmap="Greys", alpha=0.5)
cb = plt.colorbar()
cb.set_label("Free Diffusivity ($mm^2/s$)", fontsize=18)

plot_mean_with_std(ax, taus * 1e3, 1e5 * msds[2], "r", "dashdot")
plot_mean_with_std(ax, taus * 1e3, 1e5 * msds[3], "g", "dashdot")
ax.set_ylim(4, 9.5)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_title(r"Test-Retest MSD($\tau$) Subject 2", fontsize=15)
plt.savefig("qt_indices_msd.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# You can see that the MSD in both subjects increases over time, but also
# slowly levels off as time progresses. This makes sense as diffusing particles
# are becoming more restricted by surrounding tissue as time goes on. You can
# also see that for Subject 1 the index trends nearly perfectly overlap, but
# for subject 2 they are slightly off, which is also what we found in the
# paper.
#
#
#
# Next, we follow the same procedure to estimate the test-retest RTAP, RTOP and
# RTPP over diffusion time for both subject. For ease of comparison, we will
# estimate all three in the same unit [1/mm] by taking the square root of RTAP
# and the cubed root of RTOP.

# Again, first we define the data for the background illustration.
Delta_ = np.linspace(0.005, 0.02, 100)
RTXP_ = np.linspace(1, 200, 100)
Delta_grid, RTXP_grid = np.meshgrid(Delta_, RTXP_)
D_grid = 1 / (4 * RTXP_grid**2 * np.pi * Delta_grid)
D_levels = np.r_[1, 2, 3, 4, 6, 9, 15, 30] * 1e-4
D_colors = np.tile(np.linspace(0.8, 0, 7), (3, 1)).T

# We start with estimating the RTOP illustration.
fig = plt.figure(figsize=(10, 3))
ax = plt.subplot(1, 2, 1)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)

plot_mean_with_std(
    ax, taus * 1e3, rtops[0] ** (1 / 3.0), "r", "--", label="RTOP$^{1/3}$ Test"
)
plot_mean_with_std(
    ax, taus * 1e3, rtops[1] ** (1 / 3.0), "g", "--", label="RTOP$^{1/3}$ Retest"
)
ax.legend(fontsize=13)
ax.text(0.0091 * 1e3, 162, "D=3e-4", fontsize=12, rotation=-22)
ax.text(0.0091 * 1e3, 140, "D=4e-4", fontsize=12, rotation=-20)
ax.text(0.0091 * 1e3, 113, "D=6e-4", fontsize=12, rotation=-16)
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_title(r"Test-Retest RTOP($\tau$) Subject 1", fontsize=15)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_ylabel("RTOP$^{1/3}$ (1/mm)", fontsize=17)

ax = plt.subplot(1, 2, 2)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)
cb = plt.colorbar()
cb.set_label("Free Diffusivity ($mm^2/s$)", fontsize=18)

plot_mean_with_std(ax, taus * 1e3, rtops[2] ** (1 / 3.0), "r", "--")
plot_mean_with_std(ax, taus * 1e3, rtops[3] ** (1 / 3.0), "g", "--")
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_title(r"Test-Retest RTOP($\tau$) Subject 2", fontsize=15)
plt.savefig("qt_indices_rtop.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Similarly as MSD, the RTOP is related to the restriction that particles are
# experiencing and is also rotationally invariant. RTOP is defined as the
# probability that particles are found at the same position at the time of
# both gradient pulses. As time increases, the odds become smaller that a
# particle will arrive at the same position it left, which is illustrated by
# all RTOP trends in the figure. Notice that the estimated RTOP trends decrease
# less fast than free diffusion, meaning that particles experience restriction
# over time. Also notice that the RTOP trends in both subjects nearly
# perfectly overlap.
#
#
#
# Next, we estimate two directional :math:`q\tau`-indices, RTAP and RTPP,
# describing particle restriction perpendicular and parallel to the
# orientation of the principal diffusivity in that voxel. If the voxel
# describes coherent white matter (which it does in our corpus callosum
# example), then they describe properties related to restriction
# perpendicular and parallel to the axon bundles.

# First, we estimate the RTAP trends.
fig = plt.figure(figsize=(10, 3))
ax = plt.subplot(1, 2, 1)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)

plot_mean_with_std(
    ax, taus * 1e3, np.sqrt(rtaps[0]), "r", "-", label="RTAP$^{1/2}$ Test"
)
plot_mean_with_std(
    ax, taus * 1e3, np.sqrt(rtaps[1]), "g", "-", label="RTAP$^{1/2}$ Retest"
)
ax.legend(fontsize=13)
ax.text(0.0091 * 1e3, 162, "D=3e-4", fontsize=12, rotation=-22)
ax.text(0.0091 * 1e3, 140, "D=4e-4", fontsize=12, rotation=-20)
ax.text(0.0091 * 1e3, 113, "D=6e-4", fontsize=12, rotation=-16)
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_title(r"Test-Retest RTAP($\tau$) Subject 1", fontsize=15)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_ylabel("RTAP$^{1/2}$ (1/mm)", fontsize=17)

ax = plt.subplot(1, 2, 2)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)
cb = plt.colorbar()
cb.set_label("Free Diffusivity ($mm^2/s$)", fontsize=18)

plot_mean_with_std(ax, taus * 1e3, np.sqrt(rtaps[2]), "r", "-")
plot_mean_with_std(ax, taus * 1e3, np.sqrt(rtaps[3]), "g", "-")
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_title(r"Test-Retest RTAP($\tau$) Subject 2", fontsize=15)
plt.savefig("qt_indices_rtap.png")


# Finally the last one for RTPP.
fig = plt.figure(figsize=(10, 3))
ax = plt.subplot(1, 2, 1)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)

plot_mean_with_std(ax, taus * 1e3, rtpps[0], "r", ":", label="RTPP Test")
plot_mean_with_std(ax, taus * 1e3, rtpps[1], "g", ":", label="RTPP Retest")
ax.legend(fontsize=13)
ax.text(0.0091 * 1e3, 113, "D=6e-4", fontsize=12, rotation=-16)
ax.text(0.0091 * 1e3, 91, "D=9e-4", fontsize=12, rotation=-13)
ax.text(0.0091 * 1e3, 69, "D=15e-4", fontsize=12, rotation=-10)
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_title(r"Test-Retest RTPP($\tau$) Subject 1", fontsize=15)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_ylabel("RTPP (1/mm)", fontsize=17)

ax = plt.subplot(1, 2, 2)
plt.contourf(Delta_ * 1e3, RTXP_, D_grid, colors=D_colors, levels=D_levels, alpha=0.5)
cb = plt.colorbar()
cb.set_label("Free Diffusivity ($mm^2/s$)", fontsize=18)

plot_mean_with_std(ax, taus * 1e3, rtpps[2], "r", ":")
plot_mean_with_std(ax, taus * 1e3, rtpps[3], "g", ":")
ax.set_ylim(54, 170)
ax.set_xlim(0.009 * 1e3, 0.0185 * 1e3)
ax.set_xlabel("Diffusion Time (ms)", fontsize=17)
ax.set_title(r"Test-Retest RTPP($\tau$) Subject 2", fontsize=15)
plt.savefig("qt_indices_rtpp.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# As those of RTOP, the trends in RTAP and RTPP also decrease over time. It can
# be seen that RTAP$^{1/2}$ is always bigger than RTPP, which makes sense as
# particles in coherent white matter experience more restriction perpendicular
# to the white matter orientation than parallel to it. Again, in both subjects
# the test-retest RTAP and RTPP is nearly perfectly consistent.
# Aside from the estimation of :math:`q\tau`-space indices, :math:`q\tau`-dMRI
# also allows for the estimation of time-dependent ODFs. Once the Qtdmri model
# is fitted it can be simply called by qtdmri_fit.odf(sphere,
# s=sharpening_factor). This is identical to how the mapmri module functions,
# and allows one to study the time-dependence of ODF directionality.
#
#
#
# This concludes the example on qt-dMRI. As we showed, approaches such as
# qt-dMRI can help in studying the (finite-:math:`\tau`) temporal properties
# of diffusion in biological tissues. Differences in :math:`q\tau`-index trends
# could be indicative of underlying structural differences that affect the
# time-dependence of the diffusion process.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
=========================
Affine Registration in 3D
=========================

This example explains how to compute an affine transformation to register two
3D volumes by maximization of their Mutual Information :footcite:p:`Mattes2003`.
The optimization strategy is similar to that implemented in ANTS
:footcite:p:`Avants2009`.

We will do this twice. The first part of this tutorial will walk through the
details of the process with the object-oriented interface implemented in
the ``dipy.align`` module. The second part will use a simplified functional
interface.
"""

from os.path import join as pjoin

import numpy as np

from dipy.align import affine_registration, register_dwi_to_template
from dipy.align.imaffine import (
    AffineMap,
    AffineRegistration,
    MutualInformationMetric,
    transform_centers_of_mass,
)
from dipy.align.transforms import (
    AffineTransform3D,
    RigidTransform3D,
    TranslationTransform3D,
)
from dipy.data import fetch_stanford_hardi
from dipy.data.fetcher import fetch_syn_data
from dipy.io.image import load_nifti
from dipy.viz import regtools

###############################################################################
# Let's fetch two b0 volumes, the static image will be the b0 from the Stanford
# HARDI dataset

files, folder = fetch_stanford_hardi()
static_data, static_affine, static_img = load_nifti(
    pjoin(folder, "HARDI150.nii.gz"), return_img=True
)
static = np.squeeze(static_data)[..., 0]
static_grid2world = static_affine

###############################################################################
# Now the moving image

files, folder2 = fetch_syn_data()
moving_data, moving_affine, moving_img = load_nifti(
    pjoin(folder2, "b0.nii.gz"), return_img=True
)
moving = moving_data
moving_grid2world = moving_affine

###############################################################################
# We can see that the images are far from aligned by drawing one on top of
# the other. The images don't even have the same number of voxels, so in order
# to draw one on top of the other we need to resample the moving image on a
# grid of the same dimensions as the static image, we can do this by
# "transforming" the moving image using an identity transform

identity = np.eye(4)
affine_map = AffineMap(
    identity,
    domain_grid_shape=static.shape,
    domain_grid2world=static_grid2world,
    codomain_grid_shape=moving.shape,
    codomain_grid2world=moving_grid2world,
)
resampled = affine_map.transform(moving)
regtools.overlay_slices(
    static,
    resampled,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Moving",
    fname="resampled_0.png",
)
regtools.overlay_slices(
    static,
    resampled,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Moving",
    fname="resampled_1.png",
)
regtools.overlay_slices(
    static,
    resampled,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Moving",
    fname="resampled_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Input images before alignment.
#
#
# We can obtain a very rough (and fast) registration by just aligning the
# centers of mass of the two images

c_of_mass = transform_centers_of_mass(
    static,
    static_grid2world,
    moving,
    moving_grid2world,
)

###############################################################################
# We can now transform the moving image and draw it on top of the static image,
# registration is not likely to be good, but at least they will occupy roughly
# the same space

transformed = c_of_mass.transform(moving)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_com_0.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_com_1.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_com_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration result by aligning the centers of mass of the images.
#
#
# This was just a translation of the moving image towards the static image,
# now we will refine it by looking for an affine transform. We first create the
# similarity metric (Mutual Information) to be used. We need to specify the
# number of bins to be used to discretize the joint and marginal probability
# distribution functions (PDF), a typical value is 32. We also need to specify
# the percentage (an integer in (0, 100]) of voxels to be used for computing
# the PDFs, the most accurate registration will be obtained by using all
# voxels, but it is also the most time-consuming choice. We specify full
# sampling by passing None instead of an integer

nbins = 32
sampling_prop = None
metric = MutualInformationMetric(nbins=nbins, sampling_proportion=sampling_prop)

###############################################################################
# To avoid getting stuck at local optima, and to accelerate convergence, we
# use a multi-resolution strategy (similar to ANTS :footcite:p:`Avants2009`) by
# building a Gaussian Pyramid. To have as much flexibility as possible, the user
# can specify how this Gaussian Pyramid is built. First of all, we need to
# specify how many resolutions we want to use. This is indirectly specified by
# just providing a list of the number of iterations we want to perform at each
# resolution. Here we will just specify 3 resolutions and a large number of
# iterations, 10000 at the coarsest resolution, 1000 at the medium resolution
# and 100 at the finest. These are the default settings

level_iters = [10000, 1000, 100]

###############################################################################
# To compute the Gaussian pyramid, the original image is first smoothed at each
# level of the pyramid using a Gaussian kernel with the requested sigma.
# A good initial choice is [3.0, 1.0, 0.0], this is the default

sigmas = [3.0, 1.0, 0.0]

###############################################################################
# Now we specify the sub-sampling factors. A good configuration is [4, 2, 1],
# which means that, if the original image shape was (nx, ny, nz) voxels, then
# the shape of the coarsest image will be about (nx//4, ny//4, nz//4), the
# shape in the middle resolution will be about (nx//2, ny//2, nz//2) and the
# image at the finest scale has the same size as the original image. This set
# of factors is the default

factors = [4, 2, 1]

###############################################################################
# Now we go ahead and instantiate the registration class with the configuration
# we just prepared

affreg = AffineRegistration(
    metric=metric, level_iters=level_iters, sigmas=sigmas, factors=factors
)

###############################################################################
# Using AffineRegistration we can register our images in as many stages as we
# want, providing previous results as initialization for the next (the same
# logic as in ANTS). The reason why it is useful is that registration is a
# non-convex optimization problem (it may have more than one local optima),
# which means that it is very important to initialize as close to the solution
# as possible. For example, let's start with our (previously computed) rough
# transformation aligning the centers of mass of our images, and then refine it
# in three stages. First look for an optimal translation. The dictionary
# regtransforms contains all available transforms, we obtain one of them by
# providing its name and the dimension (either 2 or 3) of the image we are
# working with (since we are aligning volumes, the dimension is 3)

transform = TranslationTransform3D()
params0 = None
starting_affine = c_of_mass.affine
translation = affreg.optimize(
    static,
    moving,
    transform,
    params0,
    static_grid2world=static_grid2world,
    moving_grid2world=moving_grid2world,
    starting_affine=starting_affine,
)

###############################################################################
# If we look at the result, we can see that this translation is much better
# than simply aligning the centers of mass

transformed = translation.transform(moving)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_trans_0.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_trans_1.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_trans_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration result by translating the moving image, using Mutual
# Information.
#
#
#
# Now let's refine with a rigid transform (this may even modify our previously
# found optimal translation)

transform = RigidTransform3D()
params0 = None
starting_affine = translation.affine
rigid = affreg.optimize(
    static,
    moving,
    transform,
    params0,
    static_grid2world=static_grid2world,
    moving_grid2world=moving_grid2world,
    starting_affine=starting_affine,
)

###############################################################################
# This produces a slight rotation, and the images are now better aligned

transformed = rigid.transform(moving)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_rigid_0.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_rigid_1.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_rigid_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration result with a rigid transform, using Mutual Information.
#
#
#
# Finally, let's refine with a full affine transform (translation, rotation,
# scale and shear), it is safer to fit more degrees of freedom now since we
# must be very close to the optimal transform

transform = AffineTransform3D()
params0 = None
starting_affine = rigid.affine
affine = affreg.optimize(
    static,
    moving,
    transform,
    params0,
    static_grid2world=static_grid2world,
    moving_grid2world=moving_grid2world,
    starting_affine=starting_affine,
)

###############################################################################
# This results in a slight shear and scale

transformed = affine.transform(moving)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_affine_0.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_affine_1.png",
)
regtools.overlay_slices(
    static,
    transformed,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="transformed_affine_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration result with an affine transform, using Mutual Information.
#
#
#
# Now, let's repeat this process with a simplified functional interface.
# This interface constructs a pipeline of operations from a given list of
# transformations.

pipeline = ["center_of_mass", "translation", "rigid", "affine"]

###############################################################################
# And then applies the transformations in the pipeline on the input (from left
# to right) with a call to an `affine_registration` function, which takes
# optional settings for things like the iterations, sigmas and factors. The
# pipeline must be a list of strings with one or more of the following
# transformations: center_of_mass, translation, rigid, rigid_isoscaling,
# rigid_scaling and affine.

xformed_img, reg_affine = affine_registration(
    moving,
    static,
    moving_affine=moving_affine,
    static_affine=static_affine,
    nbins=32,
    metric="MI",
    pipeline=pipeline,
    level_iters=level_iters,
    sigmas=sigmas,
    factors=factors,
)

regtools.overlay_slices(
    static,
    xformed_img,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_affine_0.png",
)
regtools.overlay_slices(
    static,
    xformed_img,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_affine_1.png",
)
regtools.overlay_slices(
    static,
    xformed_img,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_affine_2.png",
)


###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration result with an affine transform, using functional interface.
#
#
#
#
# Alternatively, you can also use the `register_dwi_to_template` function that
# needs to also know about the gradient table of the DWI data, provided as a
# tuple of (bvals_file, bvecs_file). In this case, we are going to move the
# diffusion data to the B0 image (the opposite of the previous examples),
# which reverses what is the "moving" image and what is "static".

xformed_dwi, reg_affine = register_dwi_to_template(
    dwi=static_img,
    gtab=(pjoin(folder, "HARDI150.bval"), pjoin(folder, "HARDI150.bvec")),
    template=moving_img,
    reg_method="aff",
    nbins=32,
    metric="MI",
    pipeline=pipeline,
    level_iters=level_iters,
    sigmas=sigmas,
    factors=factors,
)

regtools.overlay_slices(
    moving,
    xformed_dwi,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_dwi_0.png",
)
regtools.overlay_slices(
    moving,
    xformed_dwi,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_dwi_1.png",
)
regtools.overlay_slices(
    moving,
    xformed_dwi,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Transformed",
    fname="xformed_dwi_2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Same again, using the `dwi_to_template` functional interface.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""A class for handling large tractography datasets.

It is built using the h5py which in turn implement
key features of the HDF5 (hierarchical data format) API [1]_.

References
----------
.. [1] http://www.hdfgroup.org/HDF5/doc/H5.intro.html
"""

import h5py
from nibabel.streamlines import ArraySequence as Streamlines
import numpy as np

from dipy.testing.decorators import warning_for_keywords

# Make sure not to carry across setup module from * import
__all__ = ["Dpy"]


class Dpy:
    @warning_for_keywords()
    def __init__(self, fname, *, mode="r", compression=0):
        """Advanced storage system for tractography based on HDF5

        Parameters
        ----------
        fname : str
            Full filename
        mode : str, optional
            Use 'r' to read, 'w' to write, and 'r+' to read and write (only if
            file already exists).
        compression : int, optional
            0 no compression to 9 maximum compression.

        Examples
        --------
        >>> import os
        >>> from tempfile import mkstemp #temp file
        >>> from dipy.io.dpy import Dpy
        >>> def dpy_example():
        ...     fd,fname = mkstemp()
        ...     fname += '.dpy'#add correct extension
        ...     dpw = Dpy(fname, mode='w')
        ...     A=np.ones((5,3))
        ...     B=2*A.copy()
        ...     C=3*A.copy()
        ...     dpw.write_track(A)
        ...     dpw.write_track(B)
        ...     dpw.write_track(C)
        ...     dpw.close()
        ...     dpr = Dpy(fname, mode='r')
        ...     dpr.read_track()
        ...     dpr.read_track()
        ...     dpr.read_tracksi([0, 1, 2, 0, 0, 2])
        ...     dpr.close()
        ...     os.remove(fname) #delete file from disk
        >>> dpy_example()
        """

        self.mode = mode
        self.f = h5py.File(fname, mode=self.mode)
        self.compression = compression

        if self.mode == "w":
            self.f.attrs["version"] = "0.0.1"

            self.streamlines = self.f.create_group("streamlines")

            self.tracks = self.streamlines.create_dataset(
                "tracks", shape=(0, 3), dtype="f4", maxshape=(None, 3), chunks=True
            )

            self.offsets = self.streamlines.create_dataset(
                "offsets", shape=(1,), dtype="i8", maxshape=(None,), chunks=True
            )

            self.curr_pos = 0
            self.offsets[:] = np.array([self.curr_pos]).astype(np.int64)

        if self.mode == "r":
            self.tracks = self.f["streamlines"]["tracks"]
            self.offsets = self.f["streamlines"]["offsets"]
            self.track_no = len(self.offsets) - 1
            self.offs_pos = 0

    def version(self):
        return self.f.attrs["version"]

    def write_track(self, track):
        """write on track each time"""
        self.tracks.resize(self.tracks.shape[0] + track.shape[0], axis=0)
        self.tracks[-track.shape[0] :] = track.astype(np.float32)
        self.curr_pos += track.shape[0]

        self.offsets.resize(self.offsets.shape[0] + 1, axis=0)
        self.offsets[-1] = self.curr_pos

    def write_tracks(self, tracks):
        """write many tracks together"""

        self.tracks.resize(self.tracks.shape[0] + tracks._data.shape[0], axis=0)
        self.tracks[-tracks._data.shape[0] :] = tracks._data

        self.offsets.resize(self.offsets.shape[0] + tracks._offsets.shape[0], axis=0)
        self.offsets[-tracks._offsets.shape[0] :] = (
            self.offsets[-tracks._offsets.shape[0] - 1]
            + tracks._offsets
            + tracks._lengths
        )

    def read_track(self):
        """read one track each time"""
        off0, off1 = self.offsets[self.offs_pos : self.offs_pos + 2]
        self.offs_pos += 1
        return self.tracks[off0:off1]

    def read_tracksi(self, indices):
        """read tracks with specific indices"""
        tracks = Streamlines()
        for i in indices:
            off0, off1 = self.offsets[i : i + 2]
            tracks.append(self.tracks[off0:off1])
        return tracks

    def read_tracks(self):
        """read the entire tractography"""
        offsets = self.offsets[:]
        TR = self.tracks[:]
        tracks = Streamlines()
        for i in range(len(offsets) - 1):
            off0, off1 = offsets[i : i + 2]
            tracks.append(TR[off0:off1])
        return tracks

    def close(self):
        self.f.close()



import warnings

import numpy as np
from scipy.optimize import linear_sum_assignment

from dipy.align.bundlemin import distance_matrix_mdf
from dipy.align.cpd import DeformableRegistration
from dipy.align.streamlinear import slr_with_qbx
from dipy.segment.clustering import QuickBundles
from dipy.segment.metricspeed import AveragePointwiseEuclideanMetric
from dipy.stats.analysis import assignment_map
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import Streamlines, length, unlist_streamlines
from dipy.utils.optpkg import optional_package
from dipy.viz.plotting import bundle_shape_profile

pd, have_pd, _ = optional_package("pandas")


def average_bundle_length(bundle):
    """Find average Euclidean length of the bundle in mm.

    Parameters
    ----------
    bundle : Streamlines
        Bundle who's average length is to be calculated.

    Returns
    -------
    int
        Average Euclidean length of bundle in mm.

    """
    metric = AveragePointwiseEuclideanMetric()
    qb = QuickBundles(threshold=85.0, metric=metric)
    clusters = qb.cluster(bundle)
    centroids = Streamlines(clusters.centroids)

    return length(centroids)[0]


def find_missing(lst, cb):
    """Find unmatched streamline indices in moving bundle.

    Parameters
    ----------
    lst : List
        List of integers containing all the streamlines indices in moving
        bundle.
    cb : List
        List of integers containing streamline indices of the moving bundle
        that were not matched to any streamline in static bundle.

    Returns
    -------
    list
        List containing unmatched streamlines from moving bundle

    """
    return [x for x in range(0, len(cb)) if x not in lst]


@warning_for_keywords()
def bundlewarp(
    static, moving, *, dist=None, alpha=0.5, beta=20, max_iter=15, affine=True
):
    """Register two bundles using nonlinear method.

     See :footcite:p:`Chandio2023` for further details about the method.

    Parameters
    ----------
    static : Streamlines
        Reference/fixed bundle.

    moving : Streamlines
        Target bundle that will be moved/registered to match the static bundle.

    dist : float, optional
        Precomputed distance matrix.

    alpha : float, optional
        Represents the trade-off between regularizing the deformation and
        having points match very closely. Lower value of alpha means high
        deformations.

    beta : int, optional
        Represents the strength of the interaction between points
        Gaussian kernel size.

    max_iter : int, optional
        Maximum number of iterations for deformation process in ml-CPD method.

    affine : boolean, optional
        If False, use rigid registration as starting point.

    Returns
    -------
    deformed_bundle : Streamlines
        Nonlinearly moved bundle (warped bundle)

    moving_aligned : Streamlines
        Linearly moved bundle (affinely moved)

    dist : np.ndarray
        Float array containing distance between moving and static bundle

    matched_pairs : np.ndarray
        Int array containing streamline correspondences between two bundles

    warp : np.ndarray
        Nonlinear warp map generated by BundleWarp

    References
    ----------
    .. footbibliography::

    """
    if alpha <= 0.01:
        warnings.warn(
            "Using alpha<=0.01 will result in extreme deformations", stacklevel=2
        )

    if average_bundle_length(static) <= 50:
        beta = 10

    x0 = "affine" if affine else "rigid"
    moving_aligned, _, _, _ = slr_with_qbx(static, moving, x0=x0, rm_small_clusters=0)

    if dist is not None:
        print("using pre-computed distances")
    else:
        dist = distance_matrix_mdf(static, moving_aligned).T

    matched_pairs = np.zeros((len(moving), 2))
    matched_pairs1 = np.asarray(linear_sum_assignment(dist)).T

    for mt in matched_pairs1:
        matched_pairs[mt[0]] = mt

    num = len(matched_pairs1)

    all_pairs = list(matched_pairs1[:, 0])
    all_matched = False

    while all_matched is False:
        num = len(all_pairs)

        if num < len(moving):
            ml = find_missing(all_pairs, moving)

            dist2 = dist[:][ml]

            # dist2 has distance among unmatched streamlines of moving bundle
            # and all static bundle's streamlines

            matched_pairs2 = np.asarray(linear_sum_assignment(dist2)).T

            for i in range(matched_pairs2.shape[0]):
                matched_pairs2[i][0] = ml[matched_pairs2[i][0]]

            for mt in matched_pairs2:
                matched_pairs[mt[0]] = mt

            all_pairs.extend(matched_pairs2[:, 0])

            num2 = num + len(matched_pairs2)
            if num2 == len(moving):
                all_matched = True
                num = num2
        else:
            all_matched = True

    deformed_bundle = Streamlines([])
    warp = []

    # Iterate over each pair of streamlines and deform them
    # Append deformed streamlines in deformed_bundle

    for _, pairs in enumerate(matched_pairs):
        s1 = static[int(pairs[1])]
        s2 = moving_aligned[int(pairs[0])]

        static_s = s1
        moving_s = s2

        reg = DeformableRegistration(
            X=static_s, Y=moving_s, alpha=alpha, beta=beta, max_iterations=max_iter
        )
        ty, pr = reg.register()
        ty = ty.astype(float)
        deformed_bundle.append(ty)
        warp.append(pr)

    warp = pd.DataFrame(warp, columns=["gaussian_kernel", "transforms"])

    # Returns deformed bundle, affinely moved bundle, distance matrix,
    # streamline correspondences, and warp field
    return deformed_bundle, moving_aligned, dist, matched_pairs, warp


def bundlewarp_vector_filed(moving_aligned, deformed_bundle):
    """Calculate vector fields.

    Vector field computation as the difference between each streamline point
    in the deformed and linearly aligned bundles

    Parameters
    ----------
    moving_aligned : Streamlines
        Linearly (affinely) moved bundle
    deformed_bundle : Streamlines
        Nonlinearly (warped) bundle

    Returns
    -------
    offsets : List
        Vector field modules
    directions : List
        Unitary vector directions
    colors : List
        Colors for bundle warping field vectors. Colors follow the convention
        used in DTI-derived maps (e.g. color FA) :footcite:p:`Pajevic1999`.

    References
    ----------
    .. footbibliography::
    """
    points_aligned, _ = unlist_streamlines(moving_aligned)
    points_deformed, _ = unlist_streamlines(deformed_bundle)
    vector_field = points_deformed - points_aligned

    offsets = np.sqrt(np.sum((vector_field) ** 2, 1))  # vector field modules

    # Normalize vectors to be unitary (directions)
    directions = vector_field / np.array([offsets]).T

    # Define colors mapping the direction vectors to RGB.
    # Absolute value generates DTI-like colors
    colors = directions

    return offsets, directions, colors


@warning_for_keywords()
def bundlewarp_shape_analysis(
    moving_aligned, deformed_bundle, *, no_disks=10, plotting=False
):
    """Calculate bundle shape difference profile.

    Bundle shape difference analysis using magnitude from BundleWarp
    displacements and BUAN.

    Depending on the number of points of a streamline, and the number of
    segments requested, multiple points may be considered for the computation
    of a given segment; a segment may contain information from a single point;
    or some segments may not contain information from any points. In the latter
    case, the segment will contain an ``np.nan`` value. The point-to-segment
    mapping is defined by the :func:`assignment_map`: for each segment index,
    the point information of the matching index positions, as returned by
    :func:`assignment_map`, are considered for the computation.

    Parameters
    ----------
    moving_aligned : Streamlines
        Linearly (affinely) moved bundle
    deformed_bundle : Streamlines
        Nonlinearly (warped) moved bundle
    no_disks : int, optional
        Number of segments to be created along the length of the bundle
    plotting : Boolean, optional
        Plot bundle shape profile

    Returns
    -------
    shape_profile : np.ndarray
        Float array containing bundlewarp displacement magnitudes along the
        length of the bundle
    stdv : np.ndarray
        Float array containing standard deviations
    """
    n = no_disks
    offsets, directions, colors = bundlewarp_vector_filed(
        moving_aligned, deformed_bundle
    )

    indx = assignment_map(deformed_bundle, deformed_bundle, n)
    indx = np.array(indx)

    rng = np.random.default_rng()

    colors = rng.random((n, 3))

    disks_color = []
    for _, ind in enumerate(indx):
        disks_color.append(tuple(colors[ind]))

    x = np.array(range(1, n + 1))
    shape_profile = np.zeros(n)
    stdv = np.zeros(n)

    for i in range(n):
        mask = indx == i
        if sum(mask):
            shape_profile[i] = np.mean(offsets[mask])
            stdv[i] = np.std(offsets[mask])
        else:
            shape_profile[i] = np.nan
            stdv[i] = np.nan

    if plotting:
        bundle_shape_profile(x, shape_profile, stdv)

    return shape_profile, stdv



"""
=======================================================
Diffeomorphic Registration with binary and fuzzy images
=======================================================

This example demonstrates registration of a binary and a fuzzy image.
This could be seen as aligning a fuzzy (sensed) image to a binary
(e.g., template) image.
"""

import matplotlib.pyplot as plt
import numpy as np
from skimage import draw, filters

from dipy.align.imwarp import SymmetricDiffeomorphicRegistration
from dipy.align.metrics import SSDMetric
from dipy.viz import regtools

###############################################################################
# Let's generate a sample template image as the combination of three ellipses.
# We will generate the fuzzy (sensed) version of the image by smoothing
# the reference image.


def draw_ellipse(img, center, axis):
    rr, cc = draw.ellipse(center[0], center[1], axis[0], axis[1], shape=img.shape)
    img[rr, cc] = 1
    return img


img_ref = np.zeros((64, 64))
img_ref = draw_ellipse(img_ref, (25, 15), (10, 5))
img_ref = draw_ellipse(img_ref, (20, 45), (15, 10))
img_ref = draw_ellipse(img_ref, (50, 40), (7, 15))

img_in = filters.gaussian(img_ref, sigma=3)

###############################################################################
# Let's define a small visualization function.


def show_images(img_ref, img_warp, fig_name):
    fig, axarr = plt.subplots(ncols=2, figsize=(12, 5))
    axarr[0].set_title("warped image & reference contour")
    axarr[0].imshow(img_warp)
    axarr[0].contour(img_ref, colors="r")
    ssd = np.sum((img_warp - img_ref) ** 2)
    axarr[1].set_title(f"difference, SSD={ssd:.02f}")
    im = axarr[1].imshow(img_warp - img_ref)
    plt.colorbar(im)
    fig.tight_layout()
    fig.savefig(fig_name + ".png")


show_images(img_ref, img_in, "input")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Input images before alignment.
#
#
#
# Let's use the general Registration function with some naive parameters,
# such as set `step_length` as 1 assuming maximal step 1 pixel and a
# reasonably small number of iterations since the deformation with already
# aligned images should be minimal.

sdr = SymmetricDiffeomorphicRegistration(
    metric=SSDMetric(img_ref.ndim),
    step_length=1.0,
    level_iters=[50, 100],
    inv_iter=50,
    ss_sigma_factor=0.1,
    opt_tol=1.0e-3,
)

###############################################################################
# Perform the registration with equal images.

mapping = sdr.optimize(img_ref.astype(float), img_ref.astype(float))
img_warp = mapping.transform(img_ref, interpolation="linear")
show_images(img_ref, img_warp, "output-0")
regtools.plot_2d_diffeomorphic_map(mapping, delta=5, fname="map-0.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration results for default parameters and equal images.
#
#
#
# Perform the registration with binary and fuzzy images.

mapping = sdr.optimize(img_ref.astype(float), img_in.astype(float))
img_warp = mapping.transform(img_in, interpolation="linear")
show_images(img_ref, img_warp, "output-1")
regtools.plot_2d_diffeomorphic_map(mapping, delta=5, fname="map-1.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration results for a naive parameter configuration.
#
#
#
# Note, we are still using a multi-scale approach which makes `step_length`
# in the upper level multiplicatively larger.
# What happens if we set `step_length` to a rather small value?

sdr.step_length = 0.1

###############################################################################
# Perform the registration and examine the output.

mapping = sdr.optimize(img_ref.astype(float), img_in.astype(float))
img_warp = mapping.transform(img_in, interpolation="linear")
show_images(img_ref, img_warp, "output-2")
regtools.plot_2d_diffeomorphic_map(mapping, delta=5, fname="map-2.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration results for decreased step size.
#
#
#
# An alternative scenario is to use just a single-scale level.
# Even though the warped image may look fine, the estimated deformations show
# that it is off the mark.

sdr = SymmetricDiffeomorphicRegistration(
    metric=SSDMetric(img_ref.ndim),
    step_length=1.0,
    level_iters=[100],
    inv_iter=50,
    ss_sigma_factor=0.1,
    opt_tol=1.0e-3,
)

###############################################################################
# Perform the registration.

mapping = sdr.optimize(img_ref.astype(float), img_in.astype(float))
img_warp = mapping.transform(img_in, interpolation="linear")
show_images(img_ref, img_warp, "output-3")
regtools.plot_2d_diffeomorphic_map(mapping, delta=5, fname="map-3.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Registration results for single level.



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

cdef inline double _apply_affine_3d_x0(double x0, double x1, double x2,
                                       double h, double[:, :] aff) nogil:
    r"""Multiplies aff by (x0, x1, x2, h), returns the 1st element of product

    Returns the first component of the product of the homogeneous matrix aff by
    (x0, x1, x2, h)
    """
    return aff[0, 0] * x0 + aff[0, 1] * x1 + aff[0, 2] * x2 + h*aff[0, 3]


cdef inline double _apply_affine_3d_x1(double x0, double x1, double x2,
                                       double h, double[:, :] aff) nogil:
    r"""Multiplies aff by (x0, x1, x2, h), returns the 2nd element of product

    Returns the first component of the product of the homogeneous matrix aff by
    (x0, x1, x2, h)
    """
    return aff[1, 0] * x0 + aff[1, 1] * x1 + aff[1, 2] * x2 + h*aff[1, 3]


cdef inline double _apply_affine_3d_x2(double x0, double x1, double x2,
                                       double h, double[:, :] aff) nogil:
    r"""Multiplies aff by (x0, x1, x2, h), returns the 3d element of product

    Returns the first component of the product of the homogeneous matrix aff by
    (x0, x1, x2, h)
    """
    return aff[2, 0] * x0 + aff[2, 1] * x1 + aff[2, 2] * x2 + h*aff[2, 3]


cdef inline double _apply_affine_2d_x0(double x0, double x1, double h,
                                       double[:, :] aff) nogil:
    r"""Multiplies aff by (x0, x1, h), returns the 1st element of product
    Returns the first component of the product of the homogeneous matrix aff by
    (x0, x1, h)
    """
    return aff[0, 0] * x0 + aff[0, 1] * x1 + h*aff[0, 2]


cdef inline double _apply_affine_2d_x1(double x0, double x1, double h,
                                       double[:, :] aff) nogil:
    r"""Multiplies aff by (x0, x1, h), returns the 2nd element of product

    Returns the first component of the product of the homogeneous matrix aff by
    (x0, x1, h)
    """
    return aff[1, 0] * x0 + aff[1, 1] * x1 + h*aff[1, 2]



import numpy as np
import numpy.testing as npt
import scipy.sparse as sps

import dipy.core.optimize as opt
from dipy.core.optimize import Optimizer, sparse_nnls, spdot
from dipy.testing.decorators import set_random_number_generator


def func(x):
    return x[0] ** 2 + x[1] ** 2 + x[2] ** 2


def func2(x):
    return x[0] ** 2 + 0.5 * x[1] ** 2 + 0.2 * x[2] ** 2 + 0.2 * x[3] ** 2


def test_optimize_new_scipy():
    opt = Optimizer(fun=func, x0=np.array([1.0, 1.0, 1.0]), method="Powell")

    npt.assert_array_almost_equal(opt.xopt, np.array([0, 0, 0]))
    npt.assert_almost_equal(opt.fopt, 0)

    opt = Optimizer(
        fun=func,
        x0=np.array([1.0, 1.0, 1.0]),
        method="L-BFGS-B",
        options={"maxcor": 10, "ftol": 1e-7, "gtol": 1e-5, "eps": 1e-8},
    )

    npt.assert_array_almost_equal(opt.xopt, np.array([0, 0, 0]))
    npt.assert_almost_equal(opt.fopt, 0)
    npt.assert_equal(opt.evolution, None)

    npt.assert_equal(opt.evolution, None)

    opt = Optimizer(
        fun=func,
        x0=np.array([1.0, 1.0, 1.0]),
        method="L-BFGS-B",
        options={"maxcor": 10, "ftol": 1e-7, "gtol": 1e-5, "eps": 1e-8},
        evolution=False,
    )

    npt.assert_array_almost_equal(opt.xopt, np.array([0, 0, 0]))
    npt.assert_almost_equal(opt.fopt, 0)

    opt.print_summary()

    opt = Optimizer(
        fun=func2,
        x0=np.array([1.0, 1.0, 1.0, 5.0]),
        method="L-BFGS-B",
        options={"maxcor": 10, "ftol": 1e-7, "gtol": 1e-5, "eps": 1e-8},
        evolution=True,
    )

    npt.assert_equal(opt.evolution.shape, (opt.nit, 4))

    opt = Optimizer(
        fun=func2,
        x0=np.array([1.0, 1.0, 1.0, 5.0]),
        method="Powell",
        options={"xtol": 1e-6, "ftol": 1e-6, "maxiter": 1e6},
        evolution=True,
    )

    npt.assert_array_almost_equal(opt.xopt, np.array([0, 0, 0, 0.0]))


@set_random_number_generator()
def test_sklearn_linear_solver(rng):
    class SillySolver(opt.SKLearnLinearSolver):
        def fit(self, X, y):
            self.coef_ = np.ones(X.shape[-1])

    MySillySolver = SillySolver()
    n_samples = 100
    n_features = 20
    y = rng.random(n_samples)
    X = np.ones((n_samples, n_features))
    MySillySolver.fit(X, y)
    npt.assert_equal(MySillySolver.coef_, np.ones(n_features))
    npt.assert_equal(MySillySolver.predict(X), np.ones(n_samples) * 20)


@set_random_number_generator()
def test_nonnegativeleastsquares(rng):
    n = 100
    X = np.eye(n)
    beta = rng.random(n)
    y = np.dot(X, beta)
    my_nnls = opt.NonNegativeLeastSquares()
    my_nnls.fit(X, y)
    npt.assert_equal(my_nnls.coef_, beta)
    npt.assert_equal(my_nnls.predict(X), y)


@set_random_number_generator()
def test_spdot(rng):
    n = 100
    m = 20
    k = 10
    A = rng.standard_normal((n, m))
    B = rng.standard_normal((m, k))
    A_sparse = sps.csr_matrix(A)
    B_sparse = sps.csr_matrix(B)
    dense_dot = np.dot(A, B)
    # Try all the different variations:
    npt.assert_array_almost_equal(dense_dot, spdot(A_sparse, B_sparse).todense())
    npt.assert_array_almost_equal(dense_dot, spdot(A, B_sparse))
    npt.assert_array_almost_equal(dense_dot, spdot(A_sparse, B))


@set_random_number_generator()
def test_sparse_nnls(rng):
    # Set up the regression:
    beta = rng.random(10)
    X = rng.standard_normal((1000, 10))
    y = np.dot(X, beta)
    beta_hat = sparse_nnls(y, X)
    beta_hat_sparse = sparse_nnls(y, sps.csr_matrix(X))
    # We should be able to get back the right answer for this simple case
    npt.assert_array_almost_equal(beta, beta_hat, decimal=1)
    npt.assert_array_almost_equal(beta, beta_hat_sparse, decimal=1)



import os.path as op
from tempfile import TemporaryDirectory

import nibabel as nib
import numpy as np
import numpy.testing as npt
import pytest

from dipy.align import (
    affine,
    affine_registration,
    center_of_mass,
    motion_correction,
    read_mapping,
    register_dwi_series,
    register_dwi_to_template,
    register_series,
    rigid,
    rigid_isoscaling,
    rigid_scaling,
    streamline_registration,
    syn_registration,
    translation,
    write_mapping,
)
from dipy.align.imwarp import DiffeomorphicMap
import dipy.core.gradients as dpg
import dipy.data as dpd
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.testing.decorators import set_random_number_generator
from dipy.tracking.utils import transform_tracking_output


def setup_module():
    global \
        subset_b0, \
        subset_dwi_data, \
        subset_t2, \
        subset_b0_img, \
        subset_t2_img, \
        gtab, \
        hardi_affine, \
        MNI_T2_affine
    MNI_T2 = dpd.read_mni_template()
    hardi_img, gtab = dpd.read_stanford_hardi()
    MNI_T2_data = MNI_T2.get_fdata()
    MNI_T2_affine = MNI_T2.affine
    hardi_data = hardi_img.get_fdata()
    hardi_affine = hardi_img.affine
    b0 = hardi_data[..., gtab.b0s_mask]
    mean_b0 = np.mean(b0, -1)

    # We select some arbitrary chunk of data so this goes quicker:
    subset_b0 = mean_b0[40:45, 40:45, 40:45]
    subset_dwi_data = nib.Nifti1Image(hardi_data[40:45, 40:45, 40:45], hardi_affine)
    subset_t2 = MNI_T2_data[40:50, 40:50, 40:50]
    subset_b0_img = nib.Nifti1Image(subset_b0, hardi_affine)
    subset_t2_img = nib.Nifti1Image(subset_t2, MNI_T2_affine)


def test_syn_registration():
    with TemporaryDirectory() as tmpdir:
        warped_moving, mapping = syn_registration(
            subset_b0,
            subset_t2,
            moving_affine=hardi_affine,
            static_affine=MNI_T2_affine,
            step_length=0.1,
            metric="CC",
            dim=3,
            level_iters=[5, 5, 5],
            sigma_diff=2.0,
            radius=1,
            prealign=None,
        )

        npt.assert_equal(warped_moving.shape, subset_t2.shape)
        mapping_fname = op.join(tmpdir, "mapping.nii.gz")
        write_mapping(mapping, mapping_fname)
        file_mapping = read_mapping(mapping_fname, subset_b0_img, subset_t2_img)

        # Test that it has the same effect on the data:
        warped_from_file = file_mapping.transform(subset_b0)
        npt.assert_equal(warped_from_file, warped_moving)

        # Test that it is, attribute by attribute, identical:
        for k in mapping.__dict__:
            npt.assert_(
                (
                    np.all(
                        mapping.__getattribute__(k) == file_mapping.__getattribute__(k)
                    )
                )
            )


def test_register_dwi_to_template():
    # Default is syn registration:
    warped_b0, mapping = register_dwi_to_template(
        subset_dwi_data,
        gtab,
        template=subset_t2_img,
        level_iters=[5, 5, 5],
        sigma_diff=2.0,
        radius=1,
    )
    npt.assert_(isinstance(mapping, DiffeomorphicMap))
    npt.assert_equal(warped_b0.shape, subset_t2_img.shape)

    # Use affine registration (+ don't provide a template and inputs as
    # strings):
    fdata, fbval, fbvec = dpd.get_fnames(name="small_64D")
    warped_data, affine_mat = register_dwi_to_template(
        fdata,
        (fbval, fbvec),
        reg_method="aff",
        level_iters=[5, 5, 5],
        sigmas=[3, 1, 0],
        factors=[4, 2, 1],
    )
    npt.assert_(isinstance(affine_mat, np.ndarray))
    npt.assert_(affine_mat.shape == (4, 4))


def test_affine_registration():
    moving = subset_b0
    static = subset_b0
    moving_affine = static_affine = np.eye(4)

    # Test without returning metric
    xformed, affine_mat = affine_registration(
        moving=moving,
        static=static,
        moving_affine=moving_affine,
        static_affine=static_affine,
        level_iters=[5, 5],
        sigmas=[3, 1],
        factors=[2, 1],
    )

    # We don't ask for much:
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # [center_of_mass] + ret_metric=True should raise an error
    with pytest.raises(ValueError):
        xformed, affine_mat = affine_registration(
            moving=moving,
            static=static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            pipeline=["center_of_mass"],
            ret_metric=True,
        )

    # Define list of methods
    reg_methods = [
        "center_of_mass",
        "translation",
        "rigid",
        "rigid_isoscaling",
        "rigid_scaling",
        "affine",
        center_of_mass,
        translation,
        rigid,
        rigid_isoscaling,
        rigid_scaling,
        affine,
    ]

    # Test methods individually (without returning any metric)
    for func in reg_methods:
        xformed, affine_mat = affine_registration(
            moving=moving,
            static=static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            level_iters=[5, 5],
            sigmas=[3, 1],
            factors=[2, 1],
            pipeline=[func],
        )
        npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # Bad method
    with pytest.raises(ValueError, match=r"^pipeline\[0\] must be one.*foo.*"):
        affine_registration(
            moving=moving,
            static=static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            pipeline=["foo"],
        )

    # Test methods individually (returning quality metric)
    expected_nparams = [0, 3, 6, 7, 9, 12] * 2
    assert len(expected_nparams) == len(reg_methods)
    for i, func in enumerate(reg_methods):
        if func in ("center_of_mass", center_of_mass):
            # can't return metric
            with pytest.raises(ValueError, match="cannot return any quality"):
                affine_registration(
                    moving=moving,
                    static=static,
                    moving_affine=moving_affine,
                    static_affine=static_affine,
                    pipeline=[func],
                    ret_metric=True,
                )
            continue

        xformed, affine_mat, xopt, fopt = affine_registration(
            moving=moving,
            static=static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            level_iters=[5, 5],
            sigmas=[3, 1],
            factors=[2, 1],
            pipeline=[func],
            ret_metric=True,
        )
        npt.assert_equal(len(xopt), expected_nparams[i])
        npt.assert_equal(isinstance(fopt, (int, float)), True)

    with pytest.raises(ValueError):
        xformed, affine_mat = affine_registration(moving, static)

    # Not supported transform names should raise an error
    npt.assert_raises(
        ValueError,
        affine_registration,
        moving,
        static,
        moving_affine=moving_affine,
        static_affine=static_affine,
        pipeline=["wrong_transform"],
    )

    # If providing nifti image objects, don't need to provide affines:
    moving_img = nib.Nifti1Image(moving, moving_affine)
    static_img = nib.Nifti1Image(static, static_affine)
    xformed, affine_mat = affine_registration(moving_img, static_img)
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # Using strings with full paths as inputs also works:
    t1_name, b0_name = dpd.get_fnames(name="syn_data")
    moving = b0_name
    static = t1_name
    xformed, affine_mat = affine_registration(
        moving, static, level_iters=[5, 5], sigmas=[3, 1], factors=[4, 2]
    )
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)


def test_single_transforms():
    moving = subset_b0
    static = subset_b0
    moving_affine = static_affine = np.eye(4)

    reg_methods = [
        center_of_mass,
        translation,
        rigid_isoscaling,
        rigid_scaling,
        rigid,
        affine,
    ]

    for func in reg_methods:
        xformed, affine_mat = func(
            moving,
            static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            level_iters=[5, 5],
            sigmas=[3, 1],
            factors=[2, 1],
        )
        # We don't ask for much:
        npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)


def test_register_series():
    fdata, fbval, fbvec = dpd.get_fnames(name="small_64D")
    img = nib.load(fdata)
    gtab = dpg.gradient_table(fbval, bvecs=fbvec)
    ref_idx = np.where(gtab.b0s_mask)[0][0]
    xformed, affines = register_series(img, ref_idx)
    npt.assert_(np.all(affines[..., ref_idx] == np.eye(4)))
    npt.assert_(np.all(xformed[..., ref_idx] == img.get_fdata()[..., ref_idx]))


def test_register_dwi_series_and_motion_correction():
    fdata, fbval, fbvec = dpd.get_fnames(name="small_64D")
    with TemporaryDirectory() as tmpdir:
        # Use an abbreviated data-set:
        img = nib.load(fdata)
        data = img.get_fdata()[..., :10]
        nib.save(nib.Nifti1Image(data, img.affine), op.join(tmpdir, "data.nii.gz"))
        # Save a subset:
        bvals = np.loadtxt(fbval)
        bvecs = np.loadtxt(fbvec)
        np.savetxt(op.join(tmpdir, "bvals.txt"), bvals[:10])
        np.savetxt(op.join(tmpdir, "bvecs.txt"), bvecs[:10])
        gtab = dpg.gradient_table(
            op.join(tmpdir, "bvals.txt"), bvecs=op.join(tmpdir, "bvecs.txt")
        )
        reg_img, reg_affines = register_dwi_series(data, gtab, affine=img.affine)
        reg_img_2, reg_affines_2 = motion_correction(data, gtab, affine=img.affine)
        npt.assert_(isinstance(reg_img, nib.Nifti1Image))

        npt.assert_array_equal(reg_img.get_fdata(), reg_img_2.get_fdata())
        npt.assert_array_equal(reg_affines, reg_affines_2)


@set_random_number_generator()
def test_streamline_registration(rng):
    sl1 = [
        np.array([[0, 0, 0], [0, 0, 0.5], [0, 0, 1], [0, 0, 1.5]]),
        np.array([[0, 0, 0], [0, 0.5, 0.5], [0, 1, 1]]),
    ]
    affine_mat = np.eye(4)
    affine_mat[:3, 3] = rng.standard_normal(3)
    sl2 = list(transform_tracking_output(sl1, affine_mat))
    aligned, matrix = streamline_registration(sl2, sl1)
    npt.assert_almost_equal(matrix, np.linalg.inv(affine_mat))
    npt.assert_almost_equal(aligned[0], sl1[0])
    npt.assert_almost_equal(aligned[1], sl1[1])

    # We assume the two tracks come from the same space, but it might have
    # some affine associated with it:
    base_aff = np.eye(4) * rng.random()
    base_aff[:3, 3] = np.array([1, 2, 3])
    base_aff[3, 3] = 1

    with TemporaryDirectory() as tmpdir:
        for use_aff in [None, base_aff]:
            fname1 = op.join(tmpdir, "sl1.trk")
            fname2 = op.join(tmpdir, "sl2.trk")
            if use_aff is not None:
                img = nib.Nifti1Image(np.zeros((2, 2, 2)), use_aff)
                # Move the streamlines to this other space, and report it:
                tgm1 = StatefulTractogram(
                    transform_tracking_output(sl1, np.linalg.inv(use_aff)),
                    img,
                    Space.VOX,
                )

                save_trk(tgm1, fname1, bbox_valid_check=False)

                tgm2 = StatefulTractogram(
                    transform_tracking_output(sl2, np.linalg.inv(use_aff)),
                    img,
                    Space.VOX,
                )

                save_trk(tgm2, fname2, bbox_valid_check=False)

            else:
                img = nib.Nifti1Image(np.zeros((2, 2, 2)), np.eye(4))
                tgm1 = StatefulTractogram(sl1, img, Space.RASMM)
                tgm2 = StatefulTractogram(sl2, img, Space.RASMM)
                save_trk(tgm1, fname1, bbox_valid_check=False)
                save_trk(tgm2, fname2, bbox_valid_check=False)

            aligned, matrix = streamline_registration(fname2, fname1)
            npt.assert_almost_equal(aligned[0], sl1[0], decimal=5)
            npt.assert_almost_equal(aligned[1], sl1[1], decimal=5)


def test_register_dwi_series_multi_b0():
    # Test if register_dwi_series works with multiple b0 images
    dwi_fname, dwi_bval_fname, dwi_bvec_fname = dpd.get_fnames(name="sherbrooke_3shell")
    data, affine = load_nifti(dwi_fname)
    bvals, bvecs = read_bvals_bvecs(dwi_bval_fname, dwi_bvec_fname)

    data_small = data[..., :3]
    data_small = np.concatenate([data[..., :1], data_small], axis=-1)
    bvals_small = np.concatenate([bvals[:1], bvals[:3]], axis=0)
    bvecs_small = np.concatenate([bvecs[:1], bvecs[:3]], axis=0)
    gtab = dpg.gradient_table(bvals_small, bvecs=bvecs_small)
    _ = motion_correction(data_small, gtab, affine=affine)



import numpy as np
from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)

import dipy.segment.featurespeed as dipysfeature
from dipy.segment.featurespeed import extract
import dipy.segment.metric as dipymetric
import dipy.segment.metricspeed as dipysmetric
from dipy.testing import assert_false, assert_true
from dipy.testing.decorators import set_random_number_generator

dtype = "float32"
rng = np.random.default_rng()
s1 = np.array([np.arange(10, dtype=dtype)] * 3).T  # 10x3
s2 = np.arange(3 * 10, dtype=dtype).reshape((-1, 3))[::-1]  # 10x3
s3 = rng.random((5, 4), dtype=dtype)  # 5x4
s4 = rng.random((5, 3), dtype=dtype)  # 5x3


def test_identity_feature():
    # Test subclassing Feature
    class IdentityFeature(dipysfeature.Feature):
        def __init__(self):
            super(IdentityFeature, self).__init__(is_order_invariant=False)

        def infer_shape(self, streamline):
            return streamline.shape

        def extract(self, streamline):
            return streamline

    for feature in [dipysfeature.IdentityFeature(), IdentityFeature()]:
        for s in [s1, s2, s3, s4]:
            # Test method infer_shape
            assert_equal(feature.infer_shape(s), s.shape)

            # Test method extract
            features = feature.extract(s)
            assert_equal(features.shape, s.shape)
            assert_array_equal(features, s)

        # This feature type is not order invariant
        assert_false(feature.is_order_invariant)
        for s in [s1, s2, s3, s4]:
            features = feature.extract(s)
            features_flip = feature.extract(s[::-1])
            assert_array_equal(features_flip, s[::-1])
            assert_true(np.any(np.not_equal(features, features_flip)))


def test_feature_resample():
    from dipy.tracking.streamline import set_number_of_points

    # Test subclassing Feature
    class ResampleFeature(dipysfeature.Feature):
        def __init__(self, nb_points):
            super(ResampleFeature, self).__init__(is_order_invariant=False)
            self.nb_points = nb_points
            if nb_points <= 0:
                msg = (
                    "ResampleFeature: `nb_points` must be strictly positive: "
                    f"{nb_points}"
                )
                raise ValueError(msg)

        def infer_shape(self, streamline):
            return self.nb_points, streamline.shape[1]

        def extract(self, streamline):
            return set_number_of_points(streamline, nb_points=self.nb_points)

    assert_raises(ValueError, dipysfeature.ResampleFeature, nb_points=0)
    assert_raises(ValueError, ResampleFeature, nb_points=0)

    max_points = max(map(len, [s1, s2, s3, s4]))
    for nb_points in [2, 5, 2 * max_points]:
        for feature in [
            dipysfeature.ResampleFeature(nb_points),
            ResampleFeature(nb_points),
        ]:
            for s in [s1, s2, s3, s4]:
                # Test method infer_shape
                assert_equal(feature.infer_shape(s), (nb_points, s.shape[1]))

                # Test method extract
                features = feature.extract(s)
                assert_equal(features.shape, (nb_points, s.shape[1]))
                assert_array_almost_equal(
                    features, set_number_of_points(s, nb_points=nb_points)
                )

            # This feature type is not order invariant
            assert_false(feature.is_order_invariant)
            for s in [s1, s2, s3, s4]:
                features = feature.extract(s)
                features_flip = feature.extract(s[::-1])
                assert_array_equal(
                    features_flip, set_number_of_points(s[::-1], nb_points=nb_points)
                )
                assert_true(np.any(np.not_equal(features, features_flip)))


def test_feature_center_of_mass():
    # Test subclassing Feature
    class CenterOfMassFeature(dipysfeature.Feature):
        def __init__(self):
            super(CenterOfMassFeature, self).__init__(is_order_invariant=True)

        def infer_shape(self, streamline):
            return 1, streamline.shape[1]

        def extract(self, streamline):
            return np.mean(streamline, axis=0)[None, :]

    for feature in [dipysfeature.CenterOfMassFeature(), CenterOfMassFeature()]:
        for s in [s1, s2, s3, s4]:
            # Test method infer_shape
            assert_equal(feature.infer_shape(s), (1, s.shape[1]))

            # Test method extract
            features = feature.extract(s)
            assert_equal(features.shape, (1, s.shape[1]))
            assert_array_almost_equal(features, np.mean(s, axis=0)[None, :])

        # This feature type is order invariant
        assert_true(feature.is_order_invariant)
        for s in [s1, s2, s3, s4]:
            features = feature.extract(s)
            features_flip = feature.extract(s[::-1])
            assert_array_almost_equal(features, features_flip)


def test_feature_midpoint():
    # Test subclassing Feature
    class MidpointFeature(dipysfeature.Feature):
        def __init__(self):
            super(MidpointFeature, self).__init__(is_order_invariant=False)

        def infer_shape(self, streamline):
            return 1, streamline.shape[1]

        def extract(self, streamline):
            return streamline[[len(streamline) // 2]]

    for feature in [dipysfeature.MidpointFeature(), MidpointFeature()]:
        for s in [s1, s2, s3, s4]:
            # Test method infer_shape
            assert_equal(feature.infer_shape(s), (1, s.shape[1]))

            # Test method extract
            features = feature.extract(s)
            assert_equal(features.shape, (1, s.shape[1]))
            assert_array_almost_equal(features, s[len(s) // 2][None, :])

        # This feature type is not order invariant
        assert_false(feature.is_order_invariant)
        for s in [s1, s2, s3, s4]:
            features = feature.extract(s)
            features_flip = feature.extract(s[::-1])
            if len(s) % 2 == 0:
                assert_true(np.any(np.not_equal(features, features_flip)))
            else:
                assert_array_equal(features, features_flip)


def test_feature_arclength():
    from dipy.tracking.streamline import length

    # Test subclassing Feature
    class ArcLengthFeature(dipysfeature.Feature):
        def __init__(self):
            super(ArcLengthFeature, self).__init__(is_order_invariant=True)

        def infer_shape(self, streamline):
            return 1, 1

        def extract(self, streamline):
            return length(streamline)[None, None]

    for feature in [dipysfeature.ArcLengthFeature(), ArcLengthFeature()]:
        for s in [s1, s2, s3, s4]:
            # Test method infer_shape
            assert_equal(feature.infer_shape(s), (1, 1))

            # Test method extract
            features = feature.extract(s)
            assert_equal(features.shape, (1, 1))
            assert_array_almost_equal(features, length(s)[None, None])

        # This feature type is order invariant
        assert_true(feature.is_order_invariant)
        for s in [s1, s2, s3, s4]:
            features = feature.extract(s)
            features_flip = feature.extract(s[::-1])
            assert_array_almost_equal(features, features_flip)


def test_feature_vector_of_endpoints():
    # Test subclassing Feature
    class VectorOfEndpointsFeature(dipysfeature.Feature):
        def __init__(self):
            super(VectorOfEndpointsFeature, self).__init__(False)

        def infer_shape(self, streamline):
            return 1, streamline.shape[1]

        def extract(self, streamline):
            return streamline[[-1]] - streamline[[0]]

    feature_types = [
        dipysfeature.VectorOfEndpointsFeature(),
        VectorOfEndpointsFeature(),
    ]
    for feature in feature_types:
        for s in [s1, s2, s3, s4]:
            # Test method infer_shape
            assert_equal(feature.infer_shape(s), (1, s.shape[1]))

            # Test method extract
            features = feature.extract(s)
            assert_equal(features.shape, (1, s.shape[1]))
            assert_array_almost_equal(features, s[[-1]] - s[[0]])

        # This feature type is not order invariant
        assert_false(feature.is_order_invariant)
        for s in [s1, s2, s3, s4]:
            features = feature.extract(s)
            features_flip = feature.extract(s[::-1])
            # The flip features are simply the negative of the features.
            assert_array_almost_equal(features, -features_flip)


@set_random_number_generator(1234)
def test_feature_extract(rng):
    # Test that features are automatically cast into float32 when
    # coming from Python space
    class CenterOfMass64bit(dipysfeature.Feature):
        def infer_shape(self, streamline):
            return streamline.shape[1]

        def extract(self, streamline):
            return np.mean(streamline.astype(np.float64), axis=0)

    nb_streamlines = 100
    feature_shape = (1, 3)  # One N-dimensional point
    feature = CenterOfMass64bit()

    nb_points = rng.integers(20, 30, size=(nb_streamlines,)) * 3
    streamlines = [
        np.arange(nb).reshape((-1, 3)).astype(np.float32) for nb in nb_points
    ]
    features = extract(feature, streamlines)

    assert_equal(len(features), len(streamlines))
    assert_equal(features[0].shape, feature_shape)

    # Test that scalar features
    class ArcLengthFeature(dipysfeature.Feature):
        def infer_shape(self, streamline):
            return 1

        def extract(self, streamline):
            square_norms = np.sum((streamline[1:] - streamline[:-1]) ** 2)
            return np.sum(np.sqrt(square_norms))

    feature_shape = (1, 1)  # One scalar represented as a 2D array
    feature = ArcLengthFeature()

    features = extract(feature, streamlines)
    assert_equal(len(features), len(streamlines))
    assert_equal(features[0].shape, feature_shape)

    # Try if streamlines are readonly
    for s in streamlines:
        s.setflags(write=False)


def test_subclassing_feature():
    class EmptyFeature(dipysfeature.Feature):
        pass

    feature = EmptyFeature()
    assert_raises(NotImplementedError, feature.infer_shape, None)
    assert_raises(NotImplementedError, feature.extract, None)


def test_using_python_feature_with_cython_metric():
    class Identity(dipysfeature.Feature):
        def infer_shape(self, streamline):
            return streamline.shape

        def extract(self, streamline):
            return streamline

    # Test using Python Feature with Cython Metric
    feature = Identity()
    metric = dipysmetric.AveragePointwiseEuclideanMetric(feature)
    d1 = dipymetric.dist(metric, s1, s2)

    features1 = metric.feature.extract(s1)
    features2 = metric.feature.extract(s2)
    d2 = metric.dist(features1, features2)
    assert_equal(d1, d2)

    # Python 2.7 on Windows 64 bits uses long type instead of int for
    # constants integer. We make sure the code is robust to such behaviour
    # by explicitly testing it.
    class ArcLengthFeature(dipysfeature.Feature):
        def infer_shape(self, streamline):
            return 1

        def extract(self, streamline):
            square_norms = np.sum((streamline[1:] - streamline[:-1]) ** 2)
            return np.sum(np.sqrt(square_norms))

    # Test using Python Feature with Cython Metric
    feature = ArcLengthFeature()
    metric = dipymetric.EuclideanMetric(feature)
    d1 = dipymetric.dist(metric, s1, s2)

    features1 = metric.feature.extract(s1)
    features2 = metric.feature.extract(s2)
    d2 = metric.dist(features1, features2)
    assert_equal(d1, d2)
